import pytest

from briefcase.config import validate_install_options_config
from briefcase.exceptions import BriefcaseConfigError


@pytest.fixture
def option_config():
    return [
        {
            "name": "first",
            "title": "First option",
            "description": "Do the first thing",
            "default": True,
        },
        {
            "name": "second",
            "title": "Second option",
            "description": "Do the second thing",
            "default": False,
        },
        # An option that is alphabetically second, but appears third,
        # and has no explicit default.
        {
            "name": "default",
            "title": "Third option",
            "description": "Do the third thing",
        },
    ]


def test_no_install_options():
    """An config can have no install options."""
    assert validate_install_options_config(None, "mystery") == {}


def test_install_options(option_config):
    """A config can provide installer options."""
    install_options = validate_install_options_config(option_config, "mystery")

    assert install_options == {
        "first": {
            "title": "First option",
            "description": "Do the first thing",
            "default": True,
        },
        "second": {
            "title": "Second option",
            "description": "Do the second thing",
            "default": False,
        },
        "default": {
            "title": "Third option",
            "description": "Do the third thing",
            "default": False,
        },
    }


def test_missing_name(option_config):
    """Install options must provide a name."""
    del option_config[1]["name"]

    with pytest.raises(
        BriefcaseConfigError,
        match=r"Mystery option 1 does not define a `name`",
    ):
        validate_install_options_config(option_config, "mystery")


def test_non_string(option_config):
    """Install option name must be a string."""
    option_config[1]["name"] = 42

    with pytest.raises(
        BriefcaseConfigError,
        match=r"Name for mystery option 1 is not a string",
    ):
        validate_install_options_config(option_config, "mystery")


@pytest.mark.parametrize(
    "name",
    [
        "1option",
        "has space",
        "has+special&chars",
    ],
)
def test_identifier_name(option_config, name):
    """Install option names can't be identifiers."""
    option_config[1]["name"] = name
    with pytest.raises(
        BriefcaseConfigError,
        match=r"' cannot be used as an mystery option name",
    ):
        validate_install_options_config(option_config, "mystery")


@pytest.mark.parametrize(
    "name",
    [
        "ALLUSERS",
        "AllUsers",
        "allusers",
    ],
)
def test_keyword_name(option_config, name):
    """Install option names can't be known keywords."""
    option_config[1]["name"] = name
    with pytest.raises(
        BriefcaseConfigError,
        match=r"' is a reserved mystery option identifier",
    ):
        validate_install_options_config(option_config, "mystery")


@pytest.mark.parametrize(
    "name",
    [
        "first",
        "FiRsT",
        "FIRST",
    ],
)
def test_unique_name(option_config, name):
    """Install option names must be unique."""
    option_config[2]["name"] = name
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Mystery option names must be unique.",
    ):
        validate_install_options_config(option_config, "mystery")


def test_missing_title(option_config):
    """Install options must provide a title."""
    del option_config[1]["title"]
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Mystery option 'second' does not provide a title.",
    ):
        validate_install_options_config(option_config, "mystery")


def test_non_string_title(option_config):
    """Install options must provide a title."""
    option_config[1]["title"] = 42

    with pytest.raises(
        BriefcaseConfigError,
        match=r"Title for mystery option 'second' is not a string.",
    ):
        validate_install_options_config(option_config, "mystery")


def test_missing_description(option_config):
    """Install options must provide a description."""
    del option_config[1]["description"]
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Mystery option 'second' does not provide a description.",
    ):
        validate_install_options_config(option_config, "mystery")


def test_non_string_description(option_config):
    """Install options must provide a string description."""
    option_config[1]["description"] = 42
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Description for mystery option 'second' is not a string.",
    ):
        validate_install_options_config(option_config, "mystery")


@pytest.mark.parametrize(
    "name",
    [
        "first",
        "FiRsT",
        "FIRST",
    ],
)
def test_duplicated_name(option_config, name):
    """Option names must be unique against other option types."""
    # Parse mystery options
    mystery_options = validate_install_options_config(option_config, "mystery")

    with pytest.raises(
        BriefcaseConfigError,
        match=(
            r"Science option names must be unique.*"
            r"already used as an mystery option"
        ),
    ):
        validate_install_options_config(
            [
                {
                    "name": "other",
                    "title": "Other option",
                    "description": "Do the other thing",
                    "default": True,
                },
                {
                    "name": name,
                    "title": "Second option",
                    "description": "Do the second thing",
                    "default": False,
                },
            ],
            "science",
            mystery=mystery_options,
        )



import os
from pathlib import Path

import pytest

from briefcase.platforms.macOS import SigningIdentity

from ....utils import create_file, create_plist_file


@pytest.fixture
def sekrit_identity():
    return SigningIdentity(id="CAFEBEEF", name="Sekrit identity (DEADBEEF)")


@pytest.fixture
def sekrit_installer_identity():
    return SigningIdentity(id="CAFEFACE", name="Sekrit Installer identity (DEADBEEF)")


@pytest.fixture
def adhoc_identity():
    return SigningIdentity()


@pytest.fixture
def first_app_with_binaries(first_app_templated, first_app_config, tmp_path):
    app_path = tmp_path / "base_path/build/first-app/macos/app/First App.app"

    # Create the stub binary
    create_file(app_path / "Contents/MacOS/First App", "Stub binary")

    # Create some libraries that need to be signed.
    lib_path = app_path / "Contents/Resources/app_packages"
    frameworks_path = app_path / "Contents/Frameworks"

    for lib in [
        "first_so.so",
        Path("subfolder/second_so.so"),
        "first_dylib.dylib",
        Path("subfolder/second_dylib.dylib"),
        "other_binary",
    ]:
        create_file(
            lib_path / lib,
            mode="wb",
            content=b"\xca\xfe\xba\xbeBinary content here",
        )

    # Mach-O file that is executable, with an odd extension
    create_file(
        lib_path / "special.binary",
        mode="wb",
        content=b"\xca\xfe\xba\xbeBinary content here",
    )
    os.chmod(lib_path / "special.binary", 0o755)

    # An embedded app
    create_file(
        lib_path / "Extras.app/Contents/MacOS/Extras",
        mode="wb",
        content=b"\xca\xfe\xba\xbeBinary content here",
    )

    # An embedded framework
    create_plist_file(frameworks_path / "Extras.framework/Resources/Info.plist", {})
    create_file(
        frameworks_path / "Extras.framework/Versions/1.2/libs/extras.dylib",
        mode="wb",
        content=b"\xca\xfe\xba\xbeBinary content here",
    )
    (frameworks_path / "Extras.framework/Versions/1.2/Extras").symlink_to(
        frameworks_path / "Extras.framework/Versions/1.2/libs/extras.dylib"
    )
    (frameworks_path / "Extras.framework/Versions/Current").symlink_to(
        frameworks_path / "Extras.framework/Versions/1.2"
    )
    (frameworks_path / "Extras.framework/Extras").symlink_to(
        frameworks_path / "Extras.framework/Versions/Current/Extras"
    )

    # Make sure there are some files in the bundle that *don't* need to be signed...
    create_file(lib_path / "first.other", "other")
    create_file(lib_path / "second.other", "other")

    # A file that has a Mach-O header, but isn't executable
    create_file(
        lib_path / "unknown.binary",
        mode="wb",
        content=b"\xca\xfe\xba\xbeother",
    )

    return first_app_config



from __future__ import annotations

import concurrent.futures
import os
import plistlib
import re
import subprocess
import time
from collections.abc import Collection
from contextlib import suppress
from pathlib import Path
from signal import SIGTERM

from packaging.version import Version

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError, NotarizationInterrupted
from briefcase.integrations.subprocess import (
    get_process_id_by_command,
    is_process_dead,
    json_parser,
)
from briefcase.integrations.xcode import XcodeCliTools, get_identities
from briefcase.platforms.macOS.filters import macOS_log_clean_filter
from briefcase.platforms.macOS.utils import AppPackagesMergeMixin, is_mach_o_binary

try:
    import dmgbuild
except ImportError:  # pragma: no-cover-if-is-macos
    # On non-macOS platforms, dmgbuild won't be installed.
    # Allow the plugin to be loaded; raise an error when tools are verified.
    dmgbuild = None


DEFAULT_OUTPUT_FORMAT = "app"

ADHOC_IDENTITY_NAME = (
    "Ad-hoc identity. The resulting package will run but cannot be re-distributed."
)


class SigningIdentity:
    def __init__(self, id="-", name=None):
        """A wrapper around the various forms of an Apple signing identity."""
        self.id = id
        if self.id == "-":
            self.team_id = None
            self.name = ADHOC_IDENTITY_NAME
            self.profile = None
        else:
            self.name = name
            self.team_id = self.team_id_from_name(name)
            self.profile = f"briefcase-macOS-{self.team_id}"

    @classmethod
    def team_id_from_name(cls, name):
        try:
            return re.match(r".*\(([\dA-Z]*)\)", name)[1]
        except TypeError:
            raise BriefcaseCommandError(
                f"Couldn't extract Team ID from signing identity {name!r}"
            ) from None

    @property
    def is_adhoc(self):
        """Is this the adhoc identity?"""
        return self.id == "-"

    def __repr__(self):
        if self.is_adhoc:
            return "<AdhocSigningIdentity>"
        else:
            return f"<SigningIdentity id={self.id}>"

    def __eq__(self, other):
        return isinstance(other, SigningIdentity) and self.id == other.id


class macOSMixin:
    platform = "macOS"
    supported_host_os: Collection[str] = {"Darwin"}
    supported_host_os_reason = "macOS applications can only be built on macOS."
    # 0.3.20 introduced a framework-based support package.
    platform_target_version: str | None = "0.3.20"

    def bundle_package_path(self, app) -> Path:
        return self.binary_path(app)

    def is_icloud_synced(self, path: Path) -> bool:
        """Determine if a path is on an iCloud drive.

        This is done by looking for the "com.apple.fileprovider.fpfs#P" resource fork.
        This fork only appears on *some* directories - most notably, `.app` folders.

        :param path: The location to check.
        :returns: True if the location has iCloud resource markers.
        """
        # Check if the path is on an iCloud mounted drive.
        try:
            # Check for the iCloud resource fork. "Good" operation produces an error,
            # so use quiet mode.
            self.tools.subprocess.check_output(
                [
                    "xattr",
                    "-p",
                    "com.apple.fileprovider.fpfs#P",
                    path,
                ],
                quiet=1,
            )
            # The resource fork was found.
            return True
        except subprocess.CalledProcessError:
            # The resource fork was not found.
            # This includes the file not existing.
            return False

    def verify_not_on_icloud(self, app: AppConfig, cleanup=False):
        """Confirm that the app is *not* on an iCloud synchronized drive.

        When a `.app` folder is on an iCloud-synchronized drive, iCloud adds filesystem
        metadata to the folder. This metadata can't be removed (iCloud will just put it
        back again), but it also conflicts with app signing. So - if we detect this
        metadata, the project has been generated somewhere that ultimately won't work.

        Optionally, this method will clean up the bundle if the verification fails.

        :param app: The app to check.
        :param cleanup: Should the app bundle be deleted if verification fails?
        """
        if self.is_icloud_synced(self.binary_path(app)):
            msg = [
                """\
Your project is in a folder that is synchronized with iCloud. This interferes
with the operation of macOS code signing."""
            ]
            if cleanup:
                self.tools.shutil.rmtree(self.bundle_path(app))
                msg.append(f"""
Move your project to a location that is not synchronized with iCloud,
and re-run `briefcase {self.command}`.""")
            else:
                bundle_path = self.bundle_path(app).relative_to(self.base_path)
                msg.append(f"""
Delete the {bundle_path} folder, move your project to location
that is not synchronized with iCloud, and re-run `briefcase {self.command}`.""")
            raise BriefcaseCommandError("\n".join(msg))


class macOSCreateMixin(AppPackagesMergeMixin):
    hidden_app_properties: Collection[str] = {"permission", "entitlement"}

    def generate_app_template(self, app: AppConfig):
        """Create an application bundle.

        :param app: The config object for the app
        """
        # Before we generate the app template, make sure the package path and formal
        # name match. This will always be the case for internal apps; they might not be
        # aligned for external apps. We can't do this in verify, because app
        # verification occurs after the template is generated.
        if self.package_path(app).name != f"{app.formal_name}.app":
            raise BriefcaseCommandError(
                "The app bundle referenced by external_package_path "
                f"({self.package_path(app).name})\n"
                f"does not match the formal name of the app ({app.formal_name!r}).\n"
            )

        super().generate_app_template(app=app)
        # If we discover we're on iCloud during app creation, we can clean up the app
        # folder. This *may* return a false negative (i.e., not accurately detect that
        # we *are* on iCloud, because it takes a moment for the iCloud daemon to detect
        # that a new folder has been created; however, if this occurs, it will be
        # picked up on the next run of any Briefcase command).
        self.verify_not_on_icloud(app, cleanup=True)

    def _install_app_requirements(
        self,
        app: AppConfig,
        requires: list[str],
        app_packages_path: Path,
        **kwargs,
    ):
        try:
            # Determine the min macOS version from the framework metadata
            # of the macos-arm64_x86_64 slice of the XCframework
            plist_file = (
                self.support_path(app)
                / "Python.xcframework/macos-arm64_x86_64"
                / "Python.framework/Resources/Info.plist"
            )
            with plist_file.open("rb") as f:
                info_plist = plistlib.load(f)

            support_min_version = info_plist.get("MinimumOSVersion", "11.0")
        except FileNotFoundError:
            # If a plist file couldn't be found, it's an old-style support package;
            # Determine min. macOS version from the VERSIONS file in the support package
            versions = dict(
                [part.strip() for part in line.split(": ", 1)]
                for line in (
                    (self.support_path(app) / "VERSIONS")
                    .read_text(encoding="UTF-8")
                    .split("\n")
                )
                if ": " in line
            )
            support_min_version = versions.get("Min macOS version", "11.0")

        # Check that the app's definition is compatible with the support package
        # If the app doesn't specify a minimum version, use the support package
        # minimum version as a default.
        macOS_min_version = getattr(app, "min_os_version", support_min_version)

        if Version(macOS_min_version) < Version(support_min_version):
            raise BriefcaseCommandError(
                "Your macOS app specifies a minimum macOS version of "
                f"{macOS_min_version}, but the support package only supports "
                f"{support_min_version}"
            )

        macOS_min_tag = macOS_min_version.replace(".", "_")

        if getattr(app, "universal_build", True):
            # Perform the initial install targeting the current platform
            host_app_packages_path = (
                self.bundle_path(app) / f"app_packages.{self.tools.host_arch}"
            )
            # A standard install, except we explicitly reject installs from source
            # tarballs with `--only-binary :all:`. This is for two reasons:
            #
            # 1. Consistency. We need to use `--only-binary :all:` when we do the second
            #    "other arch" wheel install because of use of the `--platform` argument;
            #    if we only reject source tarballs from one of the installs, then a
            #    package that only provides binary wheels for one architecture would
            #    cause inconsistent results depending on which platform was the host;
            #    and
            #
            # 2. Security. Installs from source tarball involve executing arbitrary code
            #    at time of installation; and it makes the entire development
            #    environment building the app a vector for introducing vulnerabilities
            #    into an app. Forcing the use of binary wheels ensures that we can know
            #    with certainty the provenance of any binary content in the app.
            #
            # Since Briefcase is a tool designed to produce redistributable binaries,
            # we've made the judgement call that the (minor, with known workarounds)
            # inconvenience of not being able to use source tarballs is outweighed by
            # the need to produce reliable, repeatable binary artefacts.
            super()._install_app_requirements(
                app,
                requires=requires,
                app_packages_path=host_app_packages_path,
                pip_args=[
                    "--only-binary",
                    ":all:",
                    "--platform",
                    f"macosx_{macOS_min_tag}_{self.tools.host_arch}",
                ],
                install_hint=f"""

This may be because an {self.tools.host_arch} wheel that is compatible with a minimum
macOS version of {macOS_min_version} is not available.
""",
            )

            # Install dependencies for the architecture that isn't the host architecture
            other_arch = {
                "arm64": "x86_64",
                "x86_64": "arm64",
            }[self.tools.host_arch]

            # Create a temporary folder targeting the other platform
            other_app_packages_path = (
                self.bundle_path(app) / f"app_packages.{other_arch}"
            )
            if other_app_packages_path.is_dir():
                self.tools.shutil.rmtree(other_app_packages_path)
            self.tools.os.mkdir(other_app_packages_path)

            # Find all the packages with binary components.
            # We can ignore any -universal2 packages; they're already fat.
            binary_packages = self.find_binary_packages(
                host_app_packages_path,
                universal_suffix="_universal2",
                other_suffix=f"_{other_arch}",
            )
            if binary_packages:
                with self.console.wait_bar(
                    f"Installing binary app requirements for {other_arch}..."
                ):
                    self._pip_install(
                        app,
                        app_packages_path=other_app_packages_path,
                        pip_args=[
                            "--no-deps",
                            "--platform",
                            f"macosx_{macOS_min_tag}_{other_arch}",
                            "--only-binary",
                            ":all:",
                        ]
                        + [
                            f"{package}=={version}"
                            for package, version in binary_packages
                        ],
                        install_hint=f"""

This may be because an {other_arch} wheel that is compatible with a minimum
macOS version of {macOS_min_version} is not available.

You may need to build a non-universal app by setting:

    universal_build = False

in the macOS configuration section of your pyproject.toml.
""",
                    )
            else:
                self.console.info("All packages are pure Python, or universal.")

            # If given the option of a single architecture binary or a universal2
            # binary, pip will install the single platform binary. However, a common
            # situation on macOS is for there to be an x86_64 binary and a universal2
            # binary. This means you only get a universal2 binary in the "other" install
            # pass. This then causes problems with merging, because the "other" binary
            # contains a copy of the architecture that the "host" platform provides.
            #
            # To avoid this - ensure that the libraries in the app packages for the
            # "other" arch are all thin.
            #
            # This doesn't matter if it happens the other way around - if the "host"
            # arch installs a universal binary, then the "other" arch won't be asked to
            # install a binary at all.
            self.thin_app_packages(other_app_packages_path, arch=other_arch)

            # Merge the binaries
            self.merge_app_packages(
                target_app_packages=app_packages_path,
                sources=[host_app_packages_path, other_app_packages_path],
            )
        else:
            # If we're not building a universal binary, we can do a single install pass
            # directly into the app_packages folder.
            super()._install_app_requirements(
                app,
                requires=requires,
                app_packages_path=app_packages_path,
                pip_args=[
                    "--only-binary",
                    ":all:",
                    "--platform",
                    f"macosx_{macOS_min_tag}_{self.tools.host_arch}",
                ],
            )

            # Since we're only targeting 1 architecture, we can strip any universal
            # libraries down to just the host architecture.
            self.thin_app_packages(app_packages_path, arch=self.tools.host_arch)

    def permissions_context(self, app: AppConfig, cross_platform: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param cross_platform: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        # The info.plist entries for the app
        info = {}

        # Default entitlements for all macOS apps
        entitlements = {
            "com.apple.security.cs.allow-unsigned-executable-memory": True,
            "com.apple.security.cs.disable-library-validation": True,
        }

        if cross_platform["bluetooth"]:
            entitlements["com.apple.security.device.bluetooth"] = True
            info["NSBluetoothAlwaysUsageDescription"] = cross_platform["bluetooth"]
        if cross_platform["camera"]:
            entitlements["com.apple.security.device.camera"] = True
            info["NSCameraUsageDescription"] = cross_platform["camera"]
        if cross_platform["microphone"]:
            entitlements["com.apple.security.device.microphone"] = True
            info["NSMicrophoneUsageDescription"] = cross_platform["microphone"]

        if cross_platform["background_location"]:
            info["NSLocationUsageDescription"] = cross_platform["background_location"]
            entitlements["com.apple.security.personal-information.location"] = True
        elif cross_platform["fine_location"]:
            info["NSLocationUsageDescription"] = cross_platform["fine_location"]
            entitlements["com.apple.security.personal-information.location"] = True
        elif cross_platform["coarse_location"]:
            info["NSLocationUsageDescription"] = cross_platform["coarse_location"]
            entitlements["com.apple.security.personal-information.location"] = True

        if cross_platform["photo_library"]:
            info["NSPhotoLibraryUsageDescription"] = cross_platform["photo_library"]
            entitlements["com.apple.security.personal-information.photo_library"] = True

        # Override any info and entitlement definitions
        # with the platform-specific definitions
        info.update(getattr(app, "info", {}))
        entitlements.update(getattr(app, "entitlement", {}))

        return {
            "info": info,
            "entitlements": entitlements,
        }


class macOSRunMixin:
    def run_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        """
        # Console apps must operate in non-streaming mode so that console input can
        # be handled correctly. However, if we're in test mode, we *must* stream so
        # that we can see the test exit sentinel.
        if app.console_app:
            self.run_console_app(
                app,
                passthrough=passthrough,
                **kwargs,
            )
        else:
            self.run_gui_app(
                app,
                passthrough=passthrough,
                **kwargs,
            )

    def run_console_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the console application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        """
        sub_kwargs = self._prepare_app_kwargs(app=app)
        cmdline = [self.binary_path(app) / f"Contents/MacOS/{app.formal_name}"]
        cmdline.extend(passthrough)

        if app.test_mode:
            # Stream the app's output for testing.
            # When a console app runs normally, its stdout should be connected
            # directly to the terminal to properly display the app. When its test
            # suite is running, though, Briefcase should stream the output to
            # capture the testing outcome.
            app_popen = self.tools.subprocess.Popen(
                cmdline,
                cwd=self.tools.home_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
                **sub_kwargs,
            )
            self._stream_app_logs(app, popen=app_popen)

        else:
            try:
                # Start the app directly
                self.console.info("=" * 75)
                self.tools.subprocess.run(
                    cmdline,
                    cwd=self.tools.home_path,
                    check=True,
                    stream_output=False,
                    **sub_kwargs,
                )
            except subprocess.CalledProcessError:
                # The command line app *could* return an error code, which is legal.
                # Ignore any subprocess error here.
                pass

    def run_gui_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the GUI application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        """
        # Start log stream for the app.
        # Streaming the system log is... a mess. The system log contains a
        # *lot* of noise from other processes; even if you filter by
        # process, there's a lot of macOS-generated noise. It's very
        # difficult to extract just the "user generated" stdout/err log
        # messages.
        #
        # The following sets up a log stream filter that looks for:
        #  1. a log sender that matches that app binary; or,
        #  2. a log sender of libffi, and a process that matches the app binary.
        # Case (1) works for pre-Python 3.9 static linked binaries.
        # Case (2) works for Python 3.9+ dynamic linked binaries.
        # It's not enough to filter on *just* the processImagePath,
        # as the process will generate lots of system-level messages.
        # We can't filter on *just* the senderImagePath, because other
        # apps will generate log messages that would be caught by the filter.
        sender = os.fsdecode(self.binary_path(app) / "Contents/MacOS" / app.formal_name)
        log_popen = self.tools.subprocess.Popen(
            [
                "log",
                "stream",
                "--style",
                "compact",
                "--predicate",
                (
                    f'senderImagePath=="{sender}"'
                    f' OR (processImagePath=="{sender}"'
                    ' AND senderImagePath=="/usr/lib/libffi.dylib")'
                ),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
        )

        # Wait for the log stream start up
        time.sleep(0.25)

        app_pid = None
        try:
            # Set up the log stream
            sub_kwargs = self._prepare_app_kwargs(app=app)

            # Start the app in a way that lets us stream the logs
            self.tools.subprocess.run(
                # Force a new app to be launched
                ["open", "-n", self.binary_path(app)]
                + (["--args", *passthrough] if passthrough else []),
                cwd=self.tools.home_path,
                check=True,
                **sub_kwargs,
            )

            # Find the App process ID so log streaming can exit when the app exits
            app_pid = get_process_id_by_command(
                command=str(self.binary_path(app)),
                console=self.console,
            )

            if app_pid is None:
                raise BriefcaseCommandError(
                    f"Unable to find process for app {app.app_name} "
                    "to start log streaming."
                )

            # Stream the app logs.
            self._stream_app_logs(
                app,
                popen=log_popen,
                clean_filter=macOS_log_clean_filter,
                clean_output=True,
                stop_func=lambda: is_process_dead(app_pid),
                log_stream=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(f"Unable to start app {app.app_name}.") from e
        finally:
            # Ensure the App also terminates when exiting. The ordering here is a little
            # odd; the if could/should be outside the context manager, but coverage has
            # issues with that arrangement on some Python versions (3.10, 3.14)
            with suppress(ProcessLookupError):
                if app_pid:
                    self.tools.os.kill(app_pid, SIGTERM)


class macOSSigningMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # These are abstracted to enable testing without patching.
        self.get_identities = get_identities

    def entitlements_path(self, app: AppConfig):  # pragma: no-cover-if-is-windows
        return self.bundle_path(app) / self.path_index(app, "entitlements_path")

    def select_identity(
        self,
        identity: str | None = None,
        app_identity: SigningIdentity | None = None,
        allow_adhoc: bool = True,
    ) -> SigningIdentity:
        """Get the codesigning identity to use.

        This can be either an application codesigning identity, or an installer
        identity. An installer identity must be from the same Team ID as the application
        identity.

        :param identity: A pre-specified identity (either the 40-digit hex checksum, or
            the string name of the identity). If provided, it will be validated against
            the list of available identities to confirm that it is a valid codesigning
            identity.
        :param app_identity: The application signing identity to match when producing an
            installer identity. Only non-app signing identities that match the team ID
            for the ``app_identity`` will be presented as options. Omit this value to
            select an app signing identity.
        :param allow_adhoc: Should the adhoc identities be allowed?
        :returns: The final identity to use
        """
        # If the adhoc identity is allowed, add it first so it appears first in the list
        # of options.
        identities = {}
        if allow_adhoc:
            identities["-"] = ADHOC_IDENTITY_NAME

        # Obtain the valid codesigning identities. These are the identities that could
        # be used for app signing.
        identities.update(self.get_identities(self.tools, "codesigning"))

        if app_identity:
            ident_type = "installer"
            ident_option = "--installer-identity"
            # There's no way to explicitly request a list of installer identities. As a
            # workaround, get the full, unfiltered list of all signing identities; then
            # remove any identity that:
            # 1. From a different team to the provided app identity; or
            # 2. Appears on the list of app signing identities.
            app_identities = identities
            identities = {
                key: name
                for key, name in self.get_identities(self.tools).items()
                if SigningIdentity.team_id_from_name(name) == app_identity.team_id
                and key not in app_identities
            }

            if not identities:
                raise BriefcaseCommandError(
                    "No installer signing identities for team "
                    f"{app_identity.team_id} could be found."
                )
        else:
            ident_type = "application"
            ident_option = "--identity"

        if identity:
            try:
                # Try to look up the identity as a hex checksum
                identity_name = identities[identity]
                return SigningIdentity(id=identity, name=identity_name)
            except KeyError as e:
                # Try to look up the identity as readable name
                try:
                    reverse_lookup = {name: ident for ident, name in identities.items()}
                    identity_id = reverse_lookup[identity]
                    return SigningIdentity(id=identity_id, name=identity)
                except KeyError:
                    # Not found as an ID or name
                    raise BriefcaseCommandError(
                        f"Invalid {ident_type} signing identity {identity}"
                    ) from e

        identity = self.console.selection_question(
            intro=f"Select {ident_type} signing identity to use:",
            description=f"{ident_type.title()} Signing Identity",
            options=identities,
        )
        identity_name = identities[identity]
        if identity == "-":
            self.console.info(f"""
In future, you could specify this signing identity by using:

    $ briefcase {self.command} macOS {self.output_format} --adhoc-sign ...

""")
        else:
            self.console.info(f"""
In future, you could specify this signing identity by using:

    $ briefcase {self.command} macOS {self.output_format} {ident_option} {identity} ...

or

    $ briefcase {self.command} macOS {self.output_format} {ident_option} "{identity_name}" ...

""")  # noqa: E501

        return SigningIdentity(id=identity, name=identity_name)

    def sign_file(
        self,
        path: Path,
        identity: SigningIdentity,
        entitlements: Path | None = None,
    ):
        """Code sign a file.

        :param path: The path to the file to sign.
        :param identity: The code signing identity to use.
        :param entitlements: The path to the entitlements file to use.
        """
        options = "runtime" if not identity.is_adhoc else None
        process_command = ["codesign", path, "--sign", identity.id, "--force"]

        if entitlements:
            process_command.append("--entitlements")
            process_command.append(entitlements)
        if options:
            process_command.append("--options")
            process_command.append(options)

        self.console.verbose(f"Signing {Path(path).relative_to(self.base_path)}")

        try:
            self.tools.subprocess.run(
                process_command,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            errors = e.stderr
            if any(
                msg in errors
                for msg in [
                    # File has a signature matching the Mach-O magic,
                    # but isn't actually a Mach-O binary
                    "unsupported format for signature",
                    # A folder named ``.framework`, but not actually a macOS Framework`
                    "bundle format unrecognized, invalid, or unsuitable",
                ]
            ):
                # We should not be signing this in the first place
                self.console.verbose(
                    f"... {Path(path).relative_to(self.base_path)} "
                    "does not require a signature"
                )
                return
            else:
                self.tools.subprocess.output_error(e)
                raise BriefcaseCommandError(f"Unable to code sign {path}.") from e

    def sign_app(
        self,
        app: AppConfig,
        identity: SigningIdentity,
    ):  # pragma: no-cover-if-is-windows
        """Sign an entire app with a specific identity.

        :param app: The app to sign
        :param identity: The signing identity to use
        """
        bundle_path = self.package_path(app)
        resources_path = bundle_path / "Contents/Resources"
        frameworks_path = bundle_path / "Contents/Frameworks"

        sign_targets = []

        for folder in (resources_path, frameworks_path):
            # Sign all Mach-O executable objects
            sign_targets.extend(
                path
                for path in folder.rglob("*")
                if not path.is_dir()
                and not path.is_symlink()
                and is_mach_o_binary(path)
            )

            # Sign all embedded frameworks
            sign_targets.extend(folder.rglob("*.framework"))

            # Sign all embedded app objects
            sign_targets.extend(folder.rglob("*.app"))

        # Sign the bundle path itself
        sign_targets.append(bundle_path)

        # Run signing through a ThreadPoolExecutor so that they run in parallel. We need
        # to ensure that objects are signed from the inside out (i.e., a folder must be
        # signed *after* all it's contents has been signed; files in a folder must be
        # signed *after* all subfolders in that same folder); and an app that uses a
        # library must be signed *after* all the libraries it uses. See
        # https://developer.apple.com/documentation/xcode/creating-distribution-signed-code-for-the-mac#Determine-the-signing-order
        # for details.
        #
        # To do this, we utilize grouping on a sorted list, and rely on the fact that a
        # new group is created whenever the grouping key changes. The sorting process
        # guarantees depth-first ordering; the grouping is applied over the sorted
        # content, ensuring that that folders are signed before files.
        #
        # This approach isn't perfect. It will fail if there's an embedded app that uses
        # a library in a different framework, and the app is lexically sorted before the
        # library (e.g. if app_packages/foobar/Alpha.framework/Helpers/My App.app
        # depends on app_packages/footbar/Beta.framework/libbeta). However, if you're
        # embedding apps in frameworks in Python libraries that are installed by pip,
        # you're already making poor life choices; this approach is enough to satisfy
        # the one use of app-in-framework embedding that we're aware of (PySide).
        progress_bar = self.console.progress_bar()
        task_id = progress_bar.add_task("Signing App", total=len(sign_targets))
        with progress_bar:
            for group in self.tools.file.sorted_depth_first_groups(sign_targets):
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=1 if self.console.is_deep_debug else None
                ) as executor:
                    futures = []
                    for path in group:
                        future = executor.submit(
                            self.sign_file,
                            path,
                            entitlements=self.entitlements_path(app),
                            identity=identity,
                        )
                        futures.append(future)
                    for future in concurrent.futures.as_completed(futures):
                        progress_bar.update(task_id, advance=1)
                        if future.exception():
                            raise future.exception()


class macOSPackageMixin(macOSSigningMixin):
    ADHOC_SIGN_HELP = (
        "Perform ad-hoc signing on the app. "
        "The app will only run on this machine; it cannot be redistributed to others"
    )
    IDENTITY_HELP = (
        "The code signing identity to use; either the 40-digit hex "
        "checksum, or the full name of the identity"
    )

    @property
    def packaging_formats(self):
        return ["zip", "dmg", "pkg"]

    @property
    def default_packaging_format(self):
        # The default changes depending on whether the app is a console app or a GUI app
        return None

    def notarization_path(self, app: AppConfig) -> Path:
        """The file that is submitted for notarization."""
        if app.packaging_format == "zip":
            # Notarization for bare .app's is applied to the binary, not the
            # distribution artefact, with the distribution artefact being
            # created after notarization has completed.
            return self.package_path(app)
        else:
            return self.distribution_path(app)

    def distribution_path(self, app: AppConfig) -> Path:
        """The path to the final distribution artefact."""
        if app.packaging_format == "zip":
            return self.dist_path / f"{app.formal_name}-{app.version}.app.zip"
        elif app.packaging_format == "pkg":
            return self.dist_path / f"{app.formal_name}-{app.version}.pkg"
        else:
            return self.dist_path / f"{app.formal_name}-{app.version}.dmg"

    def add_options(self, parser):
        super().add_options(parser)

        # --no-sign-installer and --installer-identity are mutually exclusive
        installer_signing_group = parser.add_mutually_exclusive_group()
        installer_signing_group.add_argument(
            "--no-sign-installer",
            dest="sign_installer",
            help="Do not sign the installer. Ignored unless using PKG format",
            action="store_false",
        )
        installer_signing_group.add_argument(
            "--installer-identity",
            dest="installer_identity",
            help=(
                "The code signing identity to use for signing the installer. "
                "Ignored unless using PKG format"
            ),
            required=False,
        )

        # We use store_const:False rather than store_false so that the
        # "unspecified" value is None, rather than True, allowing for
        # a "default behavior" interpretation when `--adhoc-sign` is specified
        parser.add_argument(
            "--no-notarize",
            dest="notarize_app",
            action="store_const",
            const=False,
            help="Disable notarization for the app",
        )
        parser.add_argument(
            "--resume",
            dest="submission_id",
            help="The notarization submission ID to resume",
            required=False,
        )

    def verify_tools(self):
        # Require the Xcode command line tools.
        XcodeCliTools.verify(tools=self.tools)

        # Verify superclass tools *after* xcode. This ensures we get the
        # git check *after* the xcode check.
        super().verify_tools()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # These are abstracted to enable testing without patching.
        self.dmgbuild = dmgbuild

    def verify_app(self, app):
        super().verify_app(app)

        if app.console_app:
            if app.packaging_format is None:
                app.packaging_format = "pkg"
            elif app.packaging_format != "pkg":
                raise BriefcaseCommandError(
                    "macOS console apps must be distributed in PKG format."
                )
        elif app.packaging_format is None:
            app.packaging_format = "dmg"

    def clean_dist_folder(self, app, **options):
        """Clean up any existing artefacts in the dist folder.

        If we are resuming a notarization session verify that the artefact exists, but
        *do not* delete it.

        :param app: The app being packaged.
        :param submission_id: The notarization submission being resumed.
        :param options: Any additional arguments passed to the package command.
        """
        if options.get("submission_id"):
            if not self.notarization_path(app).exists():
                raise BriefcaseCommandError(
                    "Notarization cannot be resumed, as the notarization artefact "
                    "associated with this app "
                    f"({self.notarization_path(app).relative_to(self.base_path)}) "
                    "does not exist."
                )

            # If the packaging format is zip, the distribution artefact is created
            # *after* completion of notarization. If there's an existing distribution
            # artefact, it must be from a previous notarization/stapling attempt.
            if app.packaging_format == "zip":
                super().clean_dist_folder(app, **options)
        else:
            super().clean_dist_folder(app, **options)

    def ditto_archive(
        self,
        app_filename: Path,
        archive_filename: Path,
    ):  # pragma: no-cover-if-not-macos
        """Create an archive of an app using ditto.

        Although the archive format is ".zip", we can't use standard Zip tools, as they
        don't preserve UTF-8 encoding on all resources. Instead, we need to use `ditto`,
        which is provided as part of macOS developer tooling. See
        https://forums.developer.apple.com/forums/thread/116831 and
        https://developer.apple.com/library/archive/technotes/tn2206/_index.html for
        more details.

        :param app_filename: The filename of the app to archive
        :param archive_filename: The filename of the archive to produce
        """
        try:
            self.tools.subprocess.run(
                [
                    "/usr/bin/ditto",
                    "-c",
                    "-k",
                    "--sequesterRsrc",
                    "--keepParent",
                    app_filename,
                    archive_filename,
                ],
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(f"Unable to archive {app_filename.name}") from e

    def notarize(
        self,
        app: AppConfig,
        identity: SigningIdentity,
        installer_identity: SigningIdentity | None = None,
    ):
        """Submit a file for notarization, and wait for that notarization to be
        completed.

        :param app: The app to notarize.
        :param identity: The code signing used to notarize the app.
        :param installer_identity: The signing identity to use when signing the
            installer. Optional unless the packaging format is ``pkg``.
        """
        # Determine the arguments that would be needed to reproduce this notarization
        if installer_identity:
            identity_args = (
                f"--identity {identity.id} --installer-identity {installer_identity.id}"
            )
            notarization_identity = installer_identity
        else:
            identity_args = f"--identity {identity.id}"
            notarization_identity = identity

        format_args = f"-p {app.packaging_format}"

        # Submit the app for notarization
        submission_id = self.submit_notarization(app, identity=notarization_identity)

        self.console.warning(f"""
Briefcase will now wait for Apple to approve the notarization request.
This can take some time - in some cases, hours.

If notarization is interrupted, you can resume by running:

    briefcase package macOS {self.output_format} {format_args} {identity_args} --resume {submission_id}

""")  # noqa: E501

        self.finalize_notarization(
            app,
            identity=notarization_identity,
            submission_id=submission_id,
        )

    def submit_notarization(self, app, identity: SigningIdentity) -> str:
        """Submit a file for notarization, returning the ID of the notarizatzion task.

        If the file is a .app, it will be archived as a .zip for submission purposes.

        :param app: The app to notarize.
        :param identity: The code signing identity to use.
        :returns: The ID of the notarization task.
        """
        filename = self.notarization_path(app)
        try:
            if app.packaging_format == "zip":
                self.console.info()
                with self.console.wait_bar(
                    f"Archiving {filename.name} for notarization..."
                ):
                    archive_filename = filename.parent / (filename.name + ".zip")
                    self.ditto_archive(filename, archive_filename)
            else:
                archive_filename = filename

            submission_id = None
            store_credentials = False
            while not submission_id:
                if store_credentials:
                    if not self.console.input_enabled:
                        raise BriefcaseCommandError(f"""
The keychain does not contain credentials for the profile {identity.profile}.
You can store these credentials by invoking:

    $ xcrun notarytool store-credentials --team-id {identity.team_id} {identity.profile}

""")

                    self.console.warning("""
The notarization process uses credentials stored on your system Keychain.
You need to do this once for each signing certificate you use.

The credentials are authenticated and stored using your Apple ID, using
an app-specific Apple ID password. To generate an app-specific Apple ID
password:

  1. Sign into https://appleid.apple.com;
  2. In the 'Sign-in and Security' section, click 'App-Specific Passwords';
  3. Click on the '+' icon. You will need to provide an identifying name
     for the password. You can pick any name that makes sense to you - the
     name is only there so you can identify passwords. 'Briefcase' would be
     one possible name.
  4. Record the password somewhere safe.
""")
                    try:
                        self.tools.subprocess.run(
                            [
                                "xcrun",
                                "notarytool",
                                "store-credentials",
                                "--team-id",
                                identity.team_id,
                                identity.profile,
                            ],
                            check=True,
                            stream_output=False,  # Command reads from stdin.
                        )
                    except subprocess.CalledProcessError as e:
                        raise BriefcaseCommandError(
                            "Unable to store credentials for team ID "
                            f"{identity.team_id}."
                        ) from e

                # Attempt the notarization
                try:
                    self.console.info()
                    with self.console.wait_bar("Submitting app for notarization..."):
                        submission = self.tools.subprocess.parse_output(
                            json_parser,
                            [
                                "xcrun",
                                "notarytool",
                                "submit",
                                archive_filename,
                                "--keychain-profile",
                                identity.profile,
                                "--output-format",
                                "json",
                            ],
                            quiet=1,
                        )
                        submission_id = submission["id"]
                except subprocess.CalledProcessError as e:
                    # Error when submitting for notarization.
                    # A return code of 69 (nice) indicates an issue with the
                    # keychain profile. If store_credentials is already True,
                    # then we've already tried to store them, so call the attempt
                    # a fail
                    if e.returncode == 69 and not store_credentials:
                        store_credentials = True
                    else:
                        self.tools.subprocess.output_error(e)
                        raise BriefcaseCommandError(
                            f"Unable to submit {filename.relative_to(self.base_path)} "
                            "for notarization."
                        ) from e
        finally:
            # If we're using .zip packaging, the archive is temporary and isn't used for
            # distribution, so we can clean up.
            if archive_filename != filename:
                self.tools.os.unlink(archive_filename)

        return submission_id

    def validate_submission_id(
        self,
        app: AppConfig,
        identity: SigningIdentity,
        submission_id: str,
    ):
        with self.console.wait_bar("Determining validity of submission ID..."):
            try:
                response = self.tools.subprocess.parse_output(
                    json_parser,
                    [
                        "xcrun",
                        "notarytool",
                        "history",
                        "--keychain-profile",
                        identity.profile,
                        "--output-format",
                        "json",
                    ],
                )

                id_matches = [
                    submission
                    for submission in response["history"]
                    if submission["id"] == submission_id
                ]

                expected_filename = id_matches[0]["name"]
                # .app files are zipped for notarization, but the app itself is
                # notarized; strip the .zip suffix for filename matching purposes.
                expected_filename = expected_filename.removesuffix(".zip")

                if expected_filename != self.notarization_path(app).name:
                    raise BriefcaseCommandError(
                        f"{submission_id} is not a submission ID for this project. "
                        f"It notarizes a file named {expected_filename}"
                    )
            except IndexError:
                raise BriefcaseCommandError(
                    f"{submission_id} is not a known submission ID for this identity."
                ) from None
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    "Unable to invoke notarytool to determine validity of "
                    "submission ID.\n"
                    "Are you sure this is the identity that was used to notarize "
                    "the app?"
                ) from e

    def finalize_notarization(
        self,
        app: AppConfig,
        identity: SigningIdentity,
        submission_id: str,
    ):
        """Finalize a notarization task.

        Polls to check the current notarization status; once notarization approval is
        received, the notarization is stapled onto the distribution artefact.

        :param app: The app to notarize.
        :param identity: The code signing identity to use.
        :param submission_id: The submission ID of the notarization task to finalize.
        """
        try:
            with self.console.wait_bar("Waiting for notarization acceptance..."):
                accepted = False
                while not accepted:
                    try:
                        response = self.tools.subprocess.parse_output(
                            json_parser,
                            [
                                "xcrun",
                                "notarytool",
                                "log",
                                "--keychain-profile",
                                identity.profile,
                                submission_id,
                            ],
                            quiet=1,
                        )

                        if response["status"] == "Accepted":
                            accepted = True
                        elif response["status"] == "Invalid":
                            summary = response.get(
                                "statusSummary", "No details provided"
                            )
                            raise BriefcaseCommandError(
                                f"Notarization was rejected: {summary}\n"
                                + "\n".join(
                                    f"""
    * ({issue.get("severity", "?")}) {issue.get("path")} [{issue.get("architecture", "unknown architecture")}]
      {issue.get("message")}
      {issue.get("docUrl", "(No additional help available)")}"""  # noqa: E501
                                    for issue in response.get("issues", [])
                                )
                            )
                        else:
                            raise BriefcaseCommandError(
                                f"Unexpected notarization status: {response['status']}"
                            )
                    except subprocess.CalledProcessError as e:
                        if e.returncode == 69:
                            # Error code 69 (nice) indicates the server can't give a log
                            # response for the provided submission ID. We've already
                            # validated that it's a valid submission ID, so that means
                            # notarization isn't complete yet. Try again in 10 seconds.
                            time.sleep(10)
                        else:
                            self.tools.subprocess.output_error(e)
                            raise BriefcaseCommandError(
                                "Unknown problem retrieving notarization status."
                            ) from e

        except KeyboardInterrupt:
            raise NotarizationInterrupted("Notarization interrupted by user.") from None
        else:
            filename = self.notarization_path(app)
            try:
                self.console.info()
                self.console.info(
                    "Stapling notarization onto "
                    f"{filename.relative_to(self.base_path)}..."
                )
                self.tools.subprocess.run(
                    ["xcrun", "stapler", "staple", filename],
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    "Unable to staple notarization onto "
                    f"{filename.relative_to(self.base_path)}"
                ) from e

        # Notarization on a zip package is performed on the bare app, so we can't
        # complete packaging until notarization has completed.
        if app.packaging_format == "zip":
            self.finalize_package_zip(app)

    def package_app(
        self,
        app: AppConfig,
        notarize_app=None,
        identity=None,
        adhoc_sign=False,
        sign_installer=True,
        installer_identity=None,
        submission_id=None,
        **kwargs,
    ):
        """Package an app bundle.

        :param app: The application to package
        :param notarize_app: Should the app be notarized? Default: ``True`` if the app
            has been signed with a real identity; ``False`` if the app is unsigned, or
            an ad-hoc signing identity has been used.
        :param identity: The signing identity to use to sign the app. This can be either
            the 40-digit hex checksum, or the string name of the identity. If
            unspecified, the user will be prompted for an app signing identity. Ignored
            if ``adhoc_sign`` is ``True``.
        :param adhoc_sign: If ``True``, code will be signed with ad-hoc identity of "-",
            and the resulting app will not be re-distributable.
        :param sign_installer: Should the installer be signed? Ignored unless the
            packaging format is ``pkg``.
        :param installer_identity: The signing identity to use when signing the
            installer. Ignored unless the packaging format is ``pkg``.
        :param submission_id: The submission ID of the notarization task to resume.
        """
        # Confirm the project isn't currently on an iCloud synced drive.
        self.verify_not_on_icloud(app)

        if submission_id:
            # If we're resuming notarization, we *can't* use an adhoc identity,
            # so don't allow it to be selected.
            identity = self.select_identity(identity=identity, allow_adhoc=False)

            if app.packaging_format == "pkg":
                notarization_identity = self.select_identity(
                    identity=installer_identity,
                    app_identity=identity,
                )
            else:
                notarization_identity = identity

            self.console.info(
                f"Resuming notarization for submission {submission_id}",
                prefix=app.app_name,
            )

            self.console.info()
            self.validate_submission_id(
                app,
                identity=notarization_identity,
                submission_id=submission_id,
            )

            self.console.info()
            self.finalize_notarization(
                app,
                identity=notarization_identity,
                submission_id=submission_id,
            )
            return

        # It's a normal packaging pass.
        self.console.info("Signing app...", prefix=app.app_name)
        if adhoc_sign:
            identity = SigningIdentity()
        else:
            identity = self.select_identity(identity=identity)

        if identity.is_adhoc:
            if notarize_app:
                raise BriefcaseCommandError(
                    "Can't notarize an app with an ad-hoc signing identity"
                )
            self.console.warning("""
*************************************************************************
** WARNING: Signing with an ad-hoc identity                            **
*************************************************************************

    This app is being signed with an ad-hoc identity. The resulting
    app will run on this computer, but will not run on anyone else's
    computer.

    To generate an app that can be distributed to others, you must
    obtain an application distribution certificate from Apple, and
    select the developer identity associated with that certificate
    when running 'briefcase package'.

*************************************************************************

""")
            self.console.info("Signing app with ad-hoc identity...")
        else:
            # If we're signing, and notarization isn't explicitly disabled,
            # notarize by default.
            if notarize_app is None:
                notarize_app = True

            self.console.info(f"Signing app with identity {identity.name}...")

        self.sign_app(app=app, identity=identity)

        if app.packaging_format == "zip":
            self.package_zip(
                app,
                notarize_app=notarize_app,
                identity=identity,
            )

        elif app.packaging_format == "pkg":
            # If the user has indicated they want to sign the installer (the default),
            # and the signing identity for the app *isn't* the adhoc identity, select an
            # identity for signing the installer.
            if sign_installer and not identity.is_adhoc:
                installer_identity = self.select_identity(
                    identity=installer_identity,
                    app_identity=identity,
                )
            else:
                installer_identity = None

            self.package_pkg(
                app,
                notarize_app=notarize_app,
                identity=identity,
                installer_identity=installer_identity,
            )

        else:  # Default packaging format is DMG
            self.package_dmg(
                app,
                notarize_app=notarize_app,
                identity=identity,
            )

    def package_zip(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
    ):
        """Package an .app bundle in a zip file.

        If we're notarizing, this doesn't actually create the distribution artefact.
        Notarization on a zip package is performed on the bare app, so we can't complete
        packaging until notarization has completed. We call ``finalize_package_zip()``
        as part of completing zip notarization.
        """
        if notarize_app:
            self.console.info(
                f"Notarizing app using team ID {identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(app, identity=identity)
        else:
            self.finalize_package_zip(app)

    def finalize_package_zip(self, app: AppConfig):
        """Finalize the zip packaging process."""
        # Build the final archive for distribution
        with self.console.wait_bar(
            f"Building final distribution archive for {self.package_path(app).name}..."
        ):
            self.ditto_archive(self.package_path(app), self.distribution_path(app))

    def package_pkg(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
        installer_identity: SigningIdentity | None,
    ):
        """Package the app as an installer."""
        dist_path: Path = self.distribution_path(app)

        self.console.info("Building PKG...", prefix=app.app_name)

        installer_path = self.bundle_path(app) / "installer"

        with self.console.wait_bar("Installing license..."):
            license_file = self.base_path / "LICENSE"
            if license_file.is_file():
                (installer_path / "resources").mkdir(exist_ok=True)
                self.tools.shutil.copy(
                    license_file,
                    installer_path / "resources/LICENSE",
                )
            else:
                raise BriefcaseCommandError("""\
Your project does not contain a LICENSE file.

Create a file named `LICENSE` in the same directory as your `pyproject.toml`
with your app's licensing terms.
""")

        # pkgbuild's default behavior is to make "relocatable" installs, which means
        # that if you've ever run the app, the installer will default to updating *that*
        # version, rather than putting it in the location that the installer specifies.
        # This means if you've ever used `briefcase run`, that will be the install
        # location of the "installed" app. To work around this, you have to provide a
        # plist file - but that requires providing a "root" folder that *only* contains
        # the products you want to install. So - we need to copy the built app to a
        # "clean" packaging location.
        with self.console.wait_bar("Copying app into products folder..."):
            installed_app_path = installer_path / "root" / self.package_path(app).name
            if installed_app_path.exists():
                self.tools.shutil.rmtree(installed_app_path)
            self.tools.shutil.copytree(
                self.package_path(app),
                installed_app_path,
                # Ensure that symlinks are preserved in the duplication.
                symlinks=True,
            )

        components_plist_path = self.bundle_path(app) / "installer/components.plist"

        with (
            self.console.wait_bar("Writing component manifest..."),
            components_plist_path.open("wb") as components_plist,
        ):
            plistlib.dump(
                [
                    {
                        "BundleHasStrictIdentifier": True,
                        "BundleIsRelocatable": False,
                        "BundleIsVersionChecked": True,
                        "BundleOverwriteAction": "upgrade",
                        "RootRelativeBundlePath": self.package_path(app).name,
                    }
                ],
                components_plist,
            )

        # Console apps are installed in /Library/Formal Name, and include the
        # post-install scripts. Normal apps are installed in /Applications, and don't
        # include the scripts.
        if app.console_app:
            install_args = [
                "--install-location",
                f"/Library/{app.formal_name}",
                "--scripts",
                installer_path / "scripts",
            ]
        else:
            install_args = ["--install-location", "/Applications"]

        with self.console.wait_bar("Building app package..."):
            installer_packages_path = installer_path / "packages"
            if installer_packages_path.exists():
                self.tools.shutil.rmtree(installer_packages_path)
            installer_packages_path.mkdir()

            self.tools.subprocess.run(
                [
                    "pkgbuild",
                    "--root",
                    installer_path / "root",
                    "--component-plist",
                    components_plist_path,
                    *install_args,
                    installer_packages_path / f"{app.app_name}.pkg",
                ],
                check=True,
            )

        # Build package
        with self.console.wait_bar(f"Building {dist_path.name}..."):
            if installer_identity:
                signing_options = ["--sign", installer_identity.id]
            else:
                signing_options = []

            self.tools.subprocess.run(
                [
                    "productbuild",
                    "--distribution",
                    installer_path / "Distribution.xml",
                    "--package-path",
                    installer_path / "packages",
                    "--resources",
                    installer_path / "resources",
                    *signing_options,
                    dist_path,
                ],
                check=True,
            )

        if notarize_app:
            self.console.info(
                f"Notarizing PKG with team ID {installer_identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(
                app,
                identity=identity,
                installer_identity=installer_identity,
            )

    def package_dmg(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
    ):
        """Package an app as a DMG installer."""
        dist_path: Path = self.distribution_path(app)
        self.console.info("Building DMG...", prefix=app.app_name)

        with self.console.wait_bar(f"Building {dist_path.name}..."):
            dmg_settings = {
                "files": [os.fsdecode(self.package_path(app))],
                "symlinks": {"Applications": "/Applications"},
                "icon_locations": {
                    self.package_path(app).name: (75, 75),
                    "Applications": (225, 75),
                },
                "window_rect": ((600, 600), (350, 150)),
                "icon_size": 64,
                "text_size": 12,
            }

            try:
                icon_filename = self.base_path / f"{app.installer_icon}.icns"
                if not icon_filename.exists():
                    self.console.warning(
                        f"Can't find {app.installer_icon}.icns "
                        "to use as DMG installer icon"
                    )
                    raise AttributeError()
            except AttributeError:
                # No installer icon specified. Fall back to the app icon
                if app.icon:
                    icon_filename = self.base_path / f"{app.icon}.icns"
                    if not icon_filename.exists():
                        self.console.warning(
                            f"Can't find {app.icon}.icns "
                            "to use as fallback DMG installer icon"
                        )
                        icon_filename = None
                else:
                    # No app icon specified either
                    icon_filename = None

            if icon_filename:
                dmg_settings["icon"] = os.fsdecode(icon_filename)

            try:
                image_filename = self.base_path / f"{app.installer_background}.png"
                if image_filename.exists():
                    dmg_settings["background"] = os.fsdecode(image_filename)
                else:
                    self.console.warning(
                        f"Can't find {app.installer_background}.png "
                        "to use as DMG background"
                    )
            except AttributeError:
                # No installer background image provided
                pass

            self.dmgbuild.build_dmg(
                filename=os.fsdecode(dist_path),
                volume_name=f"{app.formal_name} {app.version}",
                settings=dmg_settings,
            )

        self.sign_file(dist_path, identity=identity)

        if notarize_app:
            self.console.info(
                f"Notarizing DMG with team ID {identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(app, identity=identity)



import pytest

from briefcase.platforms.linux.system import find_changelog_filename

from ...utils import create_file


def test_no_changelog(tmp_path):
    """Changelog file does not exist."""
    assert find_changelog_filename(tmp_path) is None


def test_no_valid_changelog(tmp_path):
    """Changelog file exists but with invalid name format so its not accepted."""
    changelog_file = tmp_path / "base_path/invalid_changelog_format"
    create_file(changelog_file, "First App Changelog")
    assert find_changelog_filename(tmp_path / "base_path") is None


@pytest.mark.parametrize(
    "changelog_filename",
    [
        f"{name}{extension}"
        for name in ["CHANGELOG", "HISTORY", "NEWS", "RELEASES"]
        for extension in ["", ".md", ".rst", ".txt"]
    ],
)
def test_has_changelog(tmp_path, changelog_filename):
    """Makes sure all formats are accepted."""
    changelog_file = tmp_path / f"base_path/{changelog_filename}"
    create_file(changelog_file, "First App Changelog")
    assert find_changelog_filename(tmp_path / "base_path") == changelog_filename


def test_multiple_changefile(tmp_path):
    """If there's more than one changelog, only one is found."""

    changelog_file1 = tmp_path / "base_path/NEWS.txt"
    changelog_file2 = tmp_path / "base_path/CHANGELOG.md"
    create_file(changelog_file1, "First App Changelog")
    create_file(changelog_file2, "First App Changelog")
    assert find_changelog_filename(tmp_path / "base_path") == "CHANGELOG.md"



import sys
from pathlib import Path

import pytest

from briefcase.integrations.virtual_environment import VenvContext


@pytest.mark.skipif(sys.platform == "win32", reason="Unix specific test")
def test_bin_dir_unix(mock_tools, tmp_path):
    """bin_dir returns 'bin' directory on Unix systems."""
    venv_path = tmp_path / "test_venv"
    context = VenvContext(mock_tools, venv_path)

    result = context.bin_dir

    expected = venv_path / "bin"
    assert result == expected
    assert isinstance(result, Path)


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_bin_dir_windows(mock_tools, tmp_path):
    """bin_dir returns 'Scripts' directory on Windows."""
    venv_path = tmp_path / "test_venv"
    context = VenvContext(mock_tools, venv_path)

    result = context.bin_dir
    expected = venv_path / "Scripts"
    assert result == expected
    assert isinstance(result, Path)


@pytest.mark.parametrize(
    "venv_path",
    [
        Path("C:\\Users\\User\\venv"),
        Path("relative\\path\\venv"),
        Path("D:\\Projects\\myapp\\venv"),
    ],
)
@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_bin_dir_different_venv_paths(mock_tools, venv_path):
    """bin_dir works with different venv path structures."""
    context = VenvContext(mock_tools, venv_path)
    windows_result = context.bin_dir
    assert windows_result == venv_path / "Scripts"


@pytest.mark.parametrize(
    "venv_path",
    [
        Path("/home/user/venvs/myproject"),
        Path("relative/path/venv"),
        Path("/tmp/venv"),
    ],
)
@pytest.mark.skipif(sys.platform == "win32", reason="Unix specific test")
def test_bin_dir_different_venv_paths_unix(mock_tools, venv_path):
    """bin_dir works with different venv path structures."""
    context = VenvContext(mock_tools, venv_path)
    unix_result = context.bin_dir
    assert unix_result == venv_path / "bin"



from pathlib import Path
from textwrap import dedent

import pytest

from briefcase.console import Console
from briefcase.platforms.web.static import CSS_BANNER, StaticWebBuildCommand


@pytest.fixture
def build_command(tmp_path):
    return StaticWebBuildCommand(
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.fixture
def app_config(first_app_config, tmp_path):
    first_app_config._path = tmp_path / "base_path/build/first-app/web/static"
    (first_app_config._path / "www").mkdir(parents=True)
    return first_app_config


def write_target_file(base_path, rel_filename, content):
    target = base_path / "www" / rel_filename
    target.parent.mkdir(parents=True, exist_ok=True)
    target.write_text(content, encoding="utf-8")
    return target


def test_write_insert_warn_if_slot_missing(build_command, app_config, monkeypatch):
    """Warn if insert slot markers are missing from the file."""

    # File without insert markers
    file_text = "<html>No markers here</html>"
    target = write_target_file(app_config._path, "index.html", file_text)

    # Capture warnings
    warnings = []
    monkeypatch.setattr(build_command.console, "warning", warnings.append)

    # Attempt to insert into missing slot
    inserts = {"header": {"pkg": "<div/>"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)

    # Ensure warning was raised and file is unchanged
    assert any("markers not found" in w for w in warnings)
    assert target.read_text(encoding="utf-8") == file_text


def test_write_insert_warn_if_file_missing(build_command, app_config, monkeypatch):
    """Warn if the target file does not exist."""

    # Capture warnings
    warnings = []
    monkeypatch.setattr(build_command.console, "warning", warnings.append)

    # Attempt to insert into a file that doesn't exist
    inserts = {"header": {"pkg": "<div/>"}}
    build_command.write_inserts(app_config, Path("notexist.html"), inserts)

    # Ensure warning was raised
    assert any("Target notexist.html not found" in w for w in warnings)


def test_write_insert_is_idempotent(build_command, app_config):
    """Inserts should be idempotent when run multiple times."""

    # File with placeholder slot
    file_text = dedent(
        """\
        <html>
        <!--@@ header:start @@-->
        PLACEHOLDER
        <!--@@ header:end @@-->
        </html>
        """
    )
    target = write_target_file(app_config._path, "index.html", file_text)

    # Apply insert once
    inserts = {"header": {"pkg": "<p>Hello</p>"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)
    once = target.read_text(encoding="utf-8")

    # Apply insert again
    build_command.write_inserts(app_config, Path("index.html"), inserts)
    twice = target.read_text(encoding="utf-8")

    # Ensure result did not change
    assert once == twice


def test_write_insert_slot_name_regex_escaped(build_command, app_config):
    """Slots containing regex chars should be escaped and matched literally."""

    # File with marker containing regex chars
    file_text = dedent(
        """\
        <html>
        <!--@@ head[er]:start @@-->
        PLACEHOLDER
        <!--@@ head[er]:end @@-->
        </html>
        """
    )
    target = write_target_file(app_config._path, "index.html", file_text)

    # Insert into regex-like slot name
    inserts = {"head[er]": {"pkg": "<p>works</p>"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)

    # Ensure content was inserted correctly
    out = target.read_text(encoding="utf-8")
    assert "works" in out


def test_write_insert_css_packages_sorted(build_command, app_config):
    """Multiple CSS contributions should be inserted in sorted order."""

    # File with CSS marker slot
    file_text = "/*@@ css:start @@*/\nold\n/*@@ css:end @@*/\n"
    target = write_target_file(app_config._path, "static/css/briefcase.css", file_text)

    # Insert two contributions out of order
    inserts = {"css": {"b": "b{}", "a": "a{}"}}
    build_command.write_inserts(app_config, Path("static/css/briefcase.css"), inserts)

    out = target.read_text(encoding="utf-8")
    a_block = CSS_BANNER.format(package="a", content="a{}")
    b_block = CSS_BANNER.format(package="b", content="b{}")

    # Ensure alphabetical order and removal of old content
    assert out.index(a_block) < out.index(b_block)
    assert "old" not in out


def test_write_insert_replaces_all_matches(build_command, app_config):
    """All matching slots should be replaced, not just the first occurrence."""

    # File with two identical slots
    file_text = (
        "<!--@@ h:start @@-->X<!--@@ h:end @@-->\n"
        "<!--@@ h:start @@-->Y<!--@@ h:end @@-->\n"
    )
    target = write_target_file(app_config._path, "index.html", file_text)

    # Insert replacement content
    inserts = {"h": {"pkg": "Z"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)

    out = target.read_text(encoding="utf-8")

    # Both occurrences replaced
    assert "X" not in out
    assert "Y" not in out
    assert out.count("Z") == 2


def test_write_insert_handles_html_and_css_markers(build_command, app_config):
    """HTML and CSS marker styles should both be processed in one file."""

    # File containing both HTML and CSS markers
    file_text = dedent("""\
        <html>
          <!--@@ assets:start @@-->
          OLD_HTML
          <!--@@ assets:end @@-->
          <style>
          /*@@ assets:start @@*/
          OLD_CSS
          /*@@ assets:end @@*/
          </style>
        </html>
    """)
    target = write_target_file(app_config._path, "index.html", file_text)

    # Insert HTML and CSS contributions
    inserts = {"assets": {"pkgA": "<link/>", "pkgB": "h1{}"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)

    out = target.read_text(encoding="utf-8")

    # Ensure both banners appear, new content is inserted, and old content removed
    assert "<!--------------------------------------------------" in out
    assert "<link/>" in out
    assert "/**************************************************" in out
    assert "h1{}" in out
    assert "OLD_HTML" not in out
    assert "OLD_CSS" not in out


def test_write_insert_preserves_multiline_indent(build_command, app_config):
    """Inserted multiline content should preserve indentation of markers."""

    # File with indented marker slot
    file_text = "    <!--@@ header:start @@-->\n    X\n    <!--@@ header:end @@-->\n"
    target = write_target_file(app_config._path, "index.html", file_text)

    # Insert multiline content
    inserts = {"header": {"pkg": "L1\nL2"}}
    build_command.write_inserts(app_config, Path("index.html"), inserts)

    out = target.read_text(encoding="utf-8")

    # Ensure inserted lines are indented properly
    assert "\n    L1\n    L2\n" in out


def test_write_insert_ignores_empty_contributions(build_command, app_config):
    """Empty insert contributions should be ignored (no banner written)."""

    # File with CSS marker slot
    file_text = "/*@@ css:start @@*/\nX\n/*@@ css:end @@*/\n"
    target = write_target_file(app_config._path, "static/css/briefcase.css", file_text)

    # Provide empty contribution
    inserts = {"css": {"pkg": ""}}
    build_command.write_inserts(app_config, Path("static/css/briefcase.css"), inserts)

    out = target.read_text(encoding="utf-8")

    # Ensure no banner for empty contribution
    assert " * pkg" not in out



import shutil
import sys
from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import MagicMock, call

import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
    RequirementsInstallError,
    UnsupportedHostError,
)
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.iOS.xcode import iOSXcodeCreateCommand

from ....utils import create_file, create_plist_file


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return iOSXcodeCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["Linux", "Windows", "WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=r"iOS applications require Xcode, which is only available on macOS\.",
    ):
        create_command()


def test_extra_pip_args(create_command, first_app_generated, tmp_path):
    """Extra iOS-specific args are included in calls to pip during update."""
    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    create_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]

    create_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    create_command.install_app_requirements(first_app_generated)

    bundle_path = tmp_path / "base_path/build/first-app/ios/xcode"
    assert create_command.tools[first_app_generated].app_context.run.mock_calls == [
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphoneos'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "--platform=ios_12_0_arm64_iphoneos",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "Python.xcframework/ios-arm64/platform-config/arm64-iphoneos"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphonesimulator'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "--platform=ios_12_0_wonky_iphonesimulator",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "Python.xcframework/ios-arm64_x86_64-simulator/platform-config/wonky-iphonesimulator"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
    ]


@pytest.mark.parametrize("include_version", [True, False])
def test_legacy_support_format(
    create_command,
    first_app_generated,
    include_version,
    tmp_path,
):
    """A support package is in the legacy format is still supported."""
    # Remove the version tag from the framework if the test requires
    if include_version:
        version_tag = "12_0"
    else:
        version_tag = "13_0"

    # If we're testing an old config, delete the xcframework. This deletes the platform
    # config folders, forcing a fallback to the older locations.
    shutil.rmtree(
        tmp_path / "base_path/build/first-app/ios/xcode/Support/Python.xcframework"
    )
    # Create the old-style VERSIONS file.
    create_file(
        tmp_path / "base_path/build/first-app/ios/xcode/Support/VERSIONS",
        "\n".join(
            [
                "Python version: 3.10.15",
                "Build: b11",
                ("Min iOS version: 12.0" if include_version else ""),
                "---------------------",
                "BZip2: 1.0.8-1",
                "libFFI: 3.4.6-1",
                "OpenSSL: 3.0.15-1",
                "XZ: 5.6.2-1",
                "",
            ]
        ),
    )

    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    create_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]

    create_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    create_command.install_app_requirements(first_app_generated)

    bundle_path = tmp_path / "base_path/build/first-app/ios/xcode"
    assert create_command.tools[first_app_generated].app_context.run.mock_calls == [
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphoneos'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                f"--platform=ios_{version_tag}_arm64_iphoneos",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "platform-site/iphoneos.arm64"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphonesimulator'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                f"--platform=ios_{version_tag}_wonky_iphonesimulator",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "platform-site/iphonesimulator.wonky"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
    ]


def test_min_os_version(create_command, first_app_generated, tmp_path):
    """If a minimum iOS version is specified, it is used for wheel installs."""

    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    create_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]

    # Set a minimum OS version
    first_app_generated.min_os_version = "15.2"

    create_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    create_command.install_app_requirements(first_app_generated)

    bundle_path = tmp_path / "base_path/build/first-app/ios/xcode"
    assert create_command.tools[first_app_generated].app_context.run.mock_calls == [
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphoneos'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "--platform=ios_15_2_arm64_iphoneos",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "Python.xcframework/ios-arm64"
                    / "platform-config/arm64-iphoneos"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphonesimulator'}",
                "--only-binary=:all:",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "--platform=ios_15_2_wonky_iphonesimulator",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/Support"
                    / "Python.xcframework/ios-arm64_x86_64-simulator"
                    / "platform-config/wonky-iphonesimulator"
                ),
                "PIP_REQUIRE_VIRTUALENV": None,
            },
        ),
    ]


def test_framework_missing_min_version(create_command, first_app_generated, tmp_path):
    """If the iOS framework is missing a minimum version definition, raise an error."""
    # Replace the XCframework Info.plist file with a version that doesn't specify
    # a minimum iOS version.
    plist_file = (
        tmp_path
        / "base_path/build/first-app/ios/xcode/Support/Python.xcframework"
        / "ios-arm64/Python.framework/Info.plist"
    )
    plist_file.unlink()
    create_plist_file(
        plist_file,
        {
            "CFBundleSupportedPlatforms": "iPhoneOS",
            "CFBundleVersion": "3.10.15",
        },
    )

    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    create_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]

    create_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your iOS XCframework doesn't specify a minimum iOS version",
    ):
        create_command.install_app_requirements(first_app_generated)

    create_command.tools[first_app_generated].app_context.run.assert_not_called()


def test_incompatible_min_os_version(create_command, first_app_generated, tmp_path):
    """If the app's iOS version isn't compatible with the support package, an error is
    raised."""
    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    create_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]
    first_app_generated.min_os_version = "8.0"

    create_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Your iOS app specifies a minimum iOS version of 8.0, "
            r"but the support package only supports 12.0"
        ),
    ):
        create_command.install_app_requirements(first_app_generated)

    create_command.tools[first_app_generated].app_context.run.assert_not_called()


@pytest.mark.parametrize(
    ("permissions", "info", "context"),
    [
        # No permissions
        (
            {},
            {},
            {"info": {}},
        ),
        # Only custom permissions
        (
            {},
            {
                "NSCameraUsageDescription": "I need to see you",
                "NSMicrophoneUsageDescription": "I need to hear you",
            },
            {
                "info": {
                    "NSCameraUsageDescription": "I need to see you",
                    "NSMicrophoneUsageDescription": "I need to hear you",
                }
            },
        ),
        # Bluetooth permissions
        (
            {
                "bluetooth": "I need to connect to bluetooth device.",
            },
            {},
            {
                "info": {
                    "NSBluetoothAlwaysUsageDescription": "I need to connect to bluetooth device."
                },
            },
        ),
        # Camera permissions
        (
            {
                "camera": "I need to see you",
            },
            {},
            {
                "info": {
                    "NSCameraUsageDescription": "I need to see you",
                },
            },
        ),
        # Microphone permissions
        (
            {
                "microphone": "I need to hear you",
            },
            {},
            {
                "info": {
                    "NSMicrophoneUsageDescription": "I need to hear you",
                },
            },
        ),
        # Coarse location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": True,
                    "NSLocationWhenInUseUsageDescription": "I need to know roughly where you are",
                }
            },
        ),
        # Fine location permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": False,
                    "NSLocationWhenInUseUsageDescription": "I need to know exactly where you are",
                }
            },
        ),
        # Background location permissions
        (
            {
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "info": {
                    "NSLocationWhenInUseUsageDescription": "I always need to know where you are",
                    "NSLocationAlwaysAndWhenInUseUsageDescription": "I always need to know where you are",
                    "UIBackgroundModes": ["processing", "location"],
                }
            },
        ),
        # Coarse location background permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": True,
                    "NSLocationWhenInUseUsageDescription": "I need to know roughly where you are",
                    "NSLocationAlwaysAndWhenInUseUsageDescription": "I always need to know where you are",
                    "UIBackgroundModes": ["processing", "location"],
                }
            },
        ),
        # Fine location background permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": False,
                    "NSLocationWhenInUseUsageDescription": "I need to know exactly where you are",
                    "NSLocationAlwaysAndWhenInUseUsageDescription": "I always need to know where you are",
                    "UIBackgroundModes": ["processing", "location"],
                }
            },
        ),
        # Coarse and fine location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": False,
                    "NSLocationWhenInUseUsageDescription": "I need to know exactly where you are",
                }
            },
        ),
        # Coarse and fine background location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "info": {
                    "NSLocationDefaultAccuracyReduced": False,
                    "NSLocationWhenInUseUsageDescription": "I need to know exactly where you are",
                    "NSLocationAlwaysAndWhenInUseUsageDescription": "I always need to know where you are",
                    "UIBackgroundModes": ["processing", "location"],
                }
            },
        ),
        # Photo library permissions
        (
            {
                "photo_library": "I need to see your library",
            },
            {},
            {
                "info": {
                    "NSPhotoLibraryAddUsageDescription": "I need to see your library"
                }
            },
        ),
        # Override and augment by cross-platform definitions
        (
            {
                "camera": "I need to see you",
            },
            {
                "NSCameraUsageDescription": "Platform specific",
                "NSCustomPermission": "Custom message",
            },
            {
                "info": {
                    "NSCameraUsageDescription": "Platform specific",
                    "NSCustomPermission": "Custom message",
                }
            },
        ),
    ],
)
def test_permissions_context(create_command, first_app, permissions, info, context):
    """Platform-specific permissions can be added to the context."""
    # Set the permissions value
    first_app.permission = permissions
    first_app.info = info
    # Extract the cross-platform permissions
    x_permissions = create_command._x_permissions(first_app)
    # Check that the final platform permissions are rendered as expected.
    assert context == create_command.permissions_context(first_app, x_permissions)


def test_install_app_requirements_error_adds_install_hint_missing_iphoneos_wheel(
    create_command, first_app_generated
):
    """Install_hint (mentioning a missing iphoneos wheel) is added when
    RequirementsInstallError is raised by _install_app_requirements in the iOS create
    command."""
    first_app_generated.min_os_version = "15.4"
    first_app_generated.requires = ["package-one", "package_two", "package_three"]

    # Mock app_context for the generated app to simulate pip failure
    mock_app_context = MagicMock(spec=Subprocess)
    mock_app_context.run.side_effect = CalledProcessError(returncode=1, cmd="pip")
    create_command.tools[first_app_generated].app_context = mock_app_context

    # Check that _install_app_requirements raises a RequirementsInstallError with an install hint
    with pytest.raises(
        RequirementsInstallError,
        match=(
            r"This may be because the `iphoneos` wheels that are available are not compatible\n"
            r"with a minimum iOS version of 15.4."
        ),
    ):
        create_command._install_app_requirements(
            app=first_app_generated,
            requires=first_app_generated.requires,
            app_packages_path=Path("/test/path"),
        )

    # Ensure the mocked subprocess was called as expected
    mock_app_context.run.assert_called_once()


def test_install_app_requirements_error_adds_install_hint_missing_iphonesimulator_wheel(
    create_command, first_app_generated
):
    """Install_hint (mentioning a missing iphonesimulator wheel) is added when
    RequirementsInstallError is raised by _install_app_requirements in the iOS create
    command."""
    first_app_generated.min_os_version = "15.4"
    first_app_generated.requires = ["package-one", "package_two", "package_three"]

    # Mock app_context for the generated app to simulate pip failure
    mock_app_context = MagicMock(spec=Subprocess)
    mock_app_context.run.side_effect = [
        None,
        CalledProcessError(returncode=1, cmd="pip"),
    ]
    create_command.tools[first_app_generated].app_context = mock_app_context

    # Check that _install_app_requirements raises a RequirementsInstallError with an install hint
    with pytest.raises(
        RequirementsInstallError,
        match=(
            r"This may indicate that an `iphoneos` wheel could be found, but an\n"
            r"`iphonesimulator` wheel could not be found; or that the `iphonesimulator`\n"
            r"binary wheels that are available are not compatible with a minimum iOS\n"
            r"version of 15.4.\n"
        ),
    ):
        create_command._install_app_requirements(
            app=first_app_generated,
            requires=first_app_generated.requires,
            app_packages_path=Path("/test/path"),
        )

    # Ensure the mocked subprocess was called as expected
    assert mock_app_context.run.call_count == 2



import sys
from subprocess import CalledProcessError
from unittest.mock import MagicMock

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import RequirementsInstallError


@pytest.mark.parametrize("logging_level", [LogLevel.INFO, LogLevel.DEEP_DEBUG])
def test_install_requirements_no_error(dev_command, first_app, logging_level):
    """Ensure run is executed properly to install requirements."""
    # Configure logging level
    dev_command.console.verbosity = logging_level

    first_app.requires = [
        "package-one",
        "./local",
        "package_two",
    ]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
        ]
        + (["-vv"] if logging_level == LogLevel.DEEP_DEBUG else [])
        + [
            "package-one",
            "-e",
            "./local",
            "package_two",
        ],
        check=True,
        encoding="UTF-8",
    )


@pytest.mark.parametrize(
    ("requirement", "editable"),
    (
        [
            # Relative paths and single level
            ("./current-dir", True),
            ("../parent-dir", True),
            # Relative paths and multiple levels
            ("../../grandparent/package", True),
            ("./deeply/nested/package", True),
            ("../sibling/package", True),
            # Simple relative paths
            ("folder/package", True),
            ("src/mypackage", True),
            # Absolute paths
            ("/absolute/path", True),
            # SCM URLs (should NOT get -e for now):
            ("git+https://github.com/user/repo.git", False),
            ("git+ssh://git@github.com/user/repo.git", False),
            ("hg+https://bitbucket.org/user/repo", False),
            ("git+https://github.com/user/repo.git@1.2.3", False),
            ("SomeProject@git+https://github.com/user/repo.git", False),
            ("SomeProject@git+https://github.com/user/repo.git@branch", False),
            (
                "SomeProject@git+https://github.com/user/repo.git#subdirectory=sub_dir",
                False,
            ),
            # Archives (should NOT get -e):
            ("./local/package.tar.gz", False),
            ("./local/package.zip", False),
            ("./local/package.whl", False),
            ("./local/package.tar.bz2", False),
            ("./local/package.tar", False),
            # Archives specified by URL
            ("https://github.com/user/repo/archive/main.tar.gz", False),
            ("SomeProject@https://github.com/user/repo/archive/main.tar.gz", False),
            ("https://files.pythonhosted.org/packages/.../package-1.0.tar.gz", False),
            # Regular packages (should NOT get -e):
            ("package1", False),
            ("package2==1.2.3", False),
            ("package3 == 2.3.4", False),
            ("package4 >= 3.4.5", False),
        ]
        + (
            [
                # Windows paths. Windows honors unix-style paths, but on Unix,
                # windows-style paths appear as package names, so we only test
                # this option on Windows.
                ("folder\\windows", True),
                (".\\windows\\current", True),
                ("..\\windows\\parent", True),
                ("C:\\absolute\\windows", True),
            ]
            if sys.platform == "win32"
            else []
        )
    ),
)
def test_install_requirements_path_formats(
    dev_command,
    first_app,
    requirement,
    editable,
):
    """Test possible path formats that pip supports for editable installation."""
    first_app.requires = [requirement]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
        ]
        + (["-e"] if editable else [])
        + [
            requirement,
        ],
        check=True,
        encoding="UTF-8",
    )


def test_install_requirements_error(dev_command, first_app):
    """Ensure RequirementsInstallError exception is raised for install errors."""
    first_app.requires = ["package-one", "package_two", "packagethree"]

    mock_venv = MagicMock()
    mock_venv.run.side_effect = CalledProcessError(returncode=-1, cmd="pip")

    with pytest.raises(
        RequirementsInstallError,
        match=r"Unable to install requirements\.",
    ):
        dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
            "package-one",
            "package_two",
            "packagethree",
        ],
        check=True,
        encoding="UTF-8",
    )


def test_no_requirements(dev_command, first_app):
    """Ensure dependency installation is not attempted when nothing to install."""
    first_app.requires = []

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_not_called()


def test_no_requirements_with_requirement_installer_Args(dev_command, first_app):
    """Ensure dependency installation is not attempted when nothing to install, even if
    requirement installer args are defined."""
    first_app.requires = []
    first_app.requirement_installer_args = ["--no-cache"]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_not_called()


def test_install_requirements_test_mode(dev_command, first_app):
    """If an app has test requirements, they are also installed."""
    first_app.requires = ["package-one", "package_two", "packagethree"]
    first_app.test_requires = ["test-one", "test_two", "./local1"]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
            "package-one",
            "package_two",
            "packagethree",
            "test-one",
            "test_two",
            "-e",
            "./local1",
        ],
        check=True,
        encoding="UTF-8",
    )


def test_only_test_requirements(dev_command, first_app):
    """If an app only has test requirements, they're installed correctly."""
    first_app.requires = None
    first_app.test_requires = ["test-one", "test_two", "./local1"]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
            "test-one",
            "test_two",
            "-e",
            "./local1",
        ],
        check=True,
        encoding="UTF-8",
    )


def test_requirement_installer_args(dev_command, first_app):
    """If an app has requirement installer args, they're used correctly."""
    first_app.requires = ["one", "two", "./local1"]
    first_app.requirement_installer_args = ["--no-cache"]

    mock_venv = MagicMock()
    dev_command.install_dev_requirements(app=first_app, venv=mock_venv)

    mock_venv.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--upgrade",
            "one",
            "two",
            "-e",
            "./local1",
            "--no-cache",
        ],
        check=True,
        encoding="UTF-8",
    )



import pytest

from briefcase.integrations.virtual_environment import (
    NoOpEnvironment,
    VenvEnvironment,
    VirtualEnvironment,
)


@pytest.fixture
def virtual_environment(mock_tools):
    return VirtualEnvironment.verify(mock_tools)


def test_isolated(virtual_environment, venv_path):
    """Factory returns VenvEnvironment when isolated is true."""
    env = virtual_environment.create(venv_path, isolated=True, recreate=False)

    assert isinstance(env, VenvEnvironment)
    assert env.path == venv_path
    assert not env.recreate


def test_non_isolated(virtual_environment, venv_path):
    """Factory returns NoOpEnvironment when isolated is false."""
    env = virtual_environment.create(venv_path, isolated=False, recreate=False)

    assert isinstance(env, NoOpEnvironment)
    assert env.path == venv_path
    assert not env.recreate


def test_recreate_isolated(virtual_environment, venv_path):
    """An isolated environment can be re-created."""
    env = virtual_environment.create(venv_path, isolated=True, recreate=True)

    assert isinstance(env, VenvEnvironment)
    assert env.path == venv_path
    assert env.recreate


def test_recreate_non_isolated(virtual_environment, venv_path):
    """An non-isolated environment can be re-created."""
    env = virtual_environment.create(venv_path, isolated=False, recreate=True)

    assert isinstance(env, NoOpEnvironment)
    assert env.path == venv_path
    assert env.recreate



import os
import platform
import shutil
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.config import AppConfig
from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess


@pytest.fixture
def mock_tools(dummy_console, tmp_path) -> ToolCache:
    mock_tools = ToolCache(
        console=dummy_console,
        base_path=tmp_path / "tools",
        home_path=tmp_path / "home",
    )

    # Mock stdlib tools
    mock_tools.os = MagicMock(spec_set=os)
    mock_tools.platform = MagicMock(spec_set=platform)
    mock_tools.shutil = MagicMock(spec_set=shutil)
    mock_tools.sys = MagicMock(spec_set=sys)

    # Mock an empty environment
    mock_tools.os.environ = {}

    # Create base directories
    mock_tools.base_path.mkdir(parents=True)
    mock_tools.home_path.mkdir(parents=True)

    # Make File and Subprocess always available
    File.verify(tools=mock_tools)
    Subprocess.verify(tools=mock_tools)

    return mock_tools


@pytest.fixture
def first_app_config():
    return AppConfig(
        app_name="first-app",
        project_name="First Project",
        formal_name="First App",
        author="Megacorp",
        bundle="com.example",
        version="0.0.1",
        description="The first simple app",
        sources=["src/first_app"],
        license={"file": "LICENSE"},
    )



import pytest

from ....utils import create_file, create_plist_file


@pytest.fixture
def first_app_generated(first_app_config, tmp_path):
    # Create the briefcase.toml file
    create_file(
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "ios"
        / "xcode"
        / "briefcase.toml",
        "\n".join(
            [
                "[paths]",
                'app_packages_path="app_packages"',
                'support_path="Support"',
                'info_plist_path="Info.plist"',
                "",
            ],
        ),
    )

    create_plist_file(
        tmp_path / "base_path/build/first-app/ios/xcode/Info.plist",
        {
            "MainModule": "first_app",
        },
    )

    # Create the package-config folders for each platform.
    # We don't need anything in them; they just need to exist.
    xcframework_path = (
        tmp_path / "base_path/build/first-app/ios/xcode/Support/Python.xcframework"
    )

    # Create the XCframeworks's ios-arm64 Info.plist file
    # with a deliberately weird min iOS version
    create_plist_file(
        xcframework_path / "ios-arm64/Python.framework/Info.plist",
        {
            "CFBundleSupportedPlatforms": "iPhoneOS",
            "CFBundleVersion": "3.10.15",
            "MinimumOSVersion": "12.0",
        },
    )
    (xcframework_path / "ios-arm64/platform-config/arm64-iphoneos").mkdir(parents=True)

    (
        xcframework_path
        / "ios-arm64_x86_64-simulator/platform-config/arm64-iphonesimulator"
    ).mkdir(parents=True)
    (
        xcframework_path
        / "ios-arm64_x86_64-simulator/platform-config/x86_64-iphonesimulator"
    ).mkdir(parents=True)

    return first_app_config



import os

from briefcase.console import Console


def test_default_constructor(raw_console):
    """A console is enabled by default."""
    assert raw_console.input_enabled


def test_constructor_with_enabled_false():
    """A console can be constructed in a disabled state."""
    try:
        console = Console(input_enabled=False)
        assert not console.input_enabled
    finally:
        console.close()


def test_enable():
    """A disabled console can be enabled."""
    try:
        console = Console(input_enabled=False)
        assert not console.input_enabled

        console.input_enabled = True

        assert console.input_enabled
    finally:
        console.close()


def test_disable(raw_console):
    """A disabled console can be enabled."""
    assert raw_console.input_enabled

    raw_console.input_enabled = False
    assert not raw_console.input_enabled


def test_is_interactive_non_interactive(monkeypatch, raw_console):
    """Console is non-interactive when stdout has no tty."""
    monkeypatch.setattr(os, "isatty", lambda _: False)
    assert raw_console.is_interactive is False


def test_is_interactive_always_interactive(monkeypatch, raw_console):
    """Console is interactive when stdout has a tty."""
    monkeypatch.setattr(os, "isatty", lambda _: True)
    assert raw_console.is_interactive is True



from __future__ import annotations

import io
import os
import plistlib
import tarfile
import zipfile
from email.message import EmailMessage
from http import HTTPStatus
from pathlib import Path

import httpx
from httpx_retries import Retry, RetryTransport
from rich.markup import escape

from briefcase.console import Console, InputDisabled


class PartialMatchString(str):
    "A string-like class that has equality on a partial match"

    def __eq__(self, other):
        return self in other


class NoMatchString(str):
    "A string-like class that has equality when there is no match"

    def __eq__(self, other):
        return self not in other


class DummyConsole(Console):
    def __init__(self, *values, input_enabled=True):
        super().__init__(input_enabled=input_enabled)
        self.prompts = []
        self.values = list(values)

    def input(self, prompt, *args, **kwargs):
        if not self.input_enabled:
            raise InputDisabled()
        self.prompts.append(prompt)
        return self.values.pop(0)


def default_rich_prompt(prompt: str) -> str:
    """Formats a prompt as what is actually passed to Rich."""
    return f"[bold]{escape(prompt)}[/bold]"


def create_file(filepath, content, mode="w", chmod=None):
    """A test utility to create a file with known content.

    Ensures that the directory for the file exists, and writes a file with
    specific content.

    :param filepath: The path for the file to create
    :param content: A string containing the content to write.
    :param mode: The mode to open the file. This is `w` by default;
        use `wb` and provide content as a bitstring if you need to
        write a binary file.
    :param chmod: file permissions to apply
    :returns: The path to the file that was created.
    """
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with filepath.open(mode, **({} if "b" in mode else {"encoding": "utf-8"})) as f:
        f.write(content)

    if chmod:
        os.chmod(filepath, chmod)

    return filepath


def create_plist_file(plistpath, content):
    """A test utility to create a plist file with known content.

    Ensures that the directory for the file exists, and writes an XML plist with
    specific content.

    :param plistpath: The path for the plist file to create.
    :param content: A dictionary of content that plistlib can use to create the plist
        file.
    :returns: The path to the file that was created.
    """
    plistpath.parent.mkdir(parents=True, exist_ok=True)
    with plistpath.open("wb") as f:
        plistlib.dump(content, f)

    return plistpath


def create_zip_file(zippath, content):
    """A test utility to create a .zip file with known content.

    Ensures that the directory for the file exists, and writes a file with specific
    content.

    :param zippath: The path for the ZIP file to create
    :param content: A list of pairs; each pair is (path, data) describing an item to be
        added to the zip file.
    :returns: The path to the file that was created.
    """
    zippath.parent.mkdir(parents=True, exist_ok=True)
    with zipfile.ZipFile(zippath, "w") as f:
        for path, data in content:
            f.writestr(path, data=data)

    return zippath


def create_tgz_file(tgzpath, content, links=None):
    """A test utility to create a .tar.gz file with known content.

    Ensures that the directory for the file exists, and writes a file with specific
    content.

    :param tgzpath: The path for the gzipped tarfile file to create
    :param content: A list of pairs; each pair is (path, data) describing an item to be
        added to the tgz file.
    :param links: (Optional) A list of pairs; each pair is a (path, target) describing a
        symlink item to be added to the tgz file, and the file it will target.
    :returns: The path to the file that was created.
    """
    tgzpath.parent.mkdir(parents=True, exist_ok=True)
    with tarfile.open(tgzpath, "w:gz") as f:
        for path, data in content:
            tarinfo = tarfile.TarInfo(path)
            payload = data.encode("utf-8")
            tarinfo.size = len(payload)
            f.addfile(tarinfo, io.BytesIO(payload))
        if links:
            for path, target in links:
                tarinfo = tarfile.TarInfo(path)
                tarinfo.linkname = target
                tarinfo.type = tarfile.SYMTYPE
                f.addfile(tarinfo, None)

    return tgzpath


def mock_file_download(filename, content, mode="w", role=None):
    """Create a side effect function that mocks the download of a zip file.

    :param filename: The file name (*not* the path - just the file name) to
        create as a side effect
    :param content: A string containing the content to write.
    :param mode: The mode to open the file. This is `w` by default;
        use `wb` and provide content as a bitstring if you need to
        write a binary file.
    :param role: The role played by the content being downloaded
    :returns: a function that can act as a mock side effect for `file.download()`
    """

    def _download_file(url, download_path, role):
        return create_file(download_path / filename, content, mode=mode)

    return _download_file


def mock_zip_download(filename, content, role=None):
    """Create a side effect function that mocks the download of a zip file.

    :param filename: The file name (*not* the path - just the file name) to
        create as a side effect
    :param content: A string containing the content to write.
    :param role: The role played by the content being downloaded
    :returns: a function that can act as a mock side effect for `file.download()`
    """

    def _download_file(url, download_path, role):
        return create_zip_file(download_path / filename, content)

    return _download_file


def mock_tgz_download(filename, content, role=None, links=None):
    """Create a side effect function that mocks the download of a .tar.gz file.

    :param content: A string containing the content to write.
    :param role: The role played by the content being downloaded
    :param links: (Optional) A list of pairs; each pair is a (path, target) describing a
        symlink item to be added to the tgz file, and the file it will target.
    :returns: a function that can act as a mock side effect for `file.download()`
    """

    def _download_file(url, download_path, role):
        return create_tgz_file(download_path / filename, content, links)

    return _download_file


def distinfo_metadata(
    package: str = "dummy",
    version: str = "1.2.3",
    tag: str | list[str] = "py3-none-any",
):
    """Generate the content for a distinfo folder.

    :param package: The name of the package.
    :param version: The version number of the package.
    :param tag: The packaging tag (or tags) for the package.
    """
    content = []

    # INSTALLER
    installer = "pip\n"
    content.append((f"{package}-{version}.dist-info/INSTALLER", installer))

    # METADATA
    metadata = EmailMessage()
    metadata["Metadata-Version"] = "2.1"
    metadata["Name"] = package
    metadata["Version"] = version
    metadata["Summary"] = f"A packaged named {package}."
    metadata["Author-email"] = "Jane Developer <jane@example.com>"
    content.append((f"{package}-{version}.dist-info/METADATA", str(metadata)))

    # WHEEL
    wheel = EmailMessage()
    wheel["Wheel-Version"] = "1.0"
    wheel["Generator"] = "test-case"
    wheel["Root-Is-Purelib"] = "true" if tag == "py3-none-any" else "false"
    if isinstance(tag, str):
        wheel["Tag"] = tag
    else:
        for value in tag:
            wheel["Tag"] = value
    content.append((f"{package}-{version}.dist-info/WHEEL", str(wheel)))

    # RECORD
    # Create the file, but don't actually populate it.
    record = ""
    content.append((f"{package}-{version}.dist-info/RECORD", record))

    return content


def installed_package_content(
    package="dummy",
    version="1.2.3",
    tag="py3-none-any",
    extra_content=None,
):
    """Generate the content for an installed package.

    :param path: The site-packages folder into which to install the package.
    :param package: The name of the package in the wheel. Defaults to ``dummy``
    :param version: The version number of the package. Defaults to ``1.2.3``
    :param tag: The installation tag for the package. Defaults to a pure python wheel.
    :param extra_content: Optional. A list of tuples of ``(path, content)`` that will be
        added to the wheel.
    """
    return (
        [
            (f"{package}/__init__.py", ""),
            (f"{package}/app.py", "# This is the app"),
        ]
        + (extra_content or [])
        + distinfo_metadata(package=package, version=version, tag=tag)
    )


def create_installed_package(
    path,
    package="dummy",
    version="1.2.3",
    tag="py3-none-any",
    extra_content=None,
):
    """Write an installed package into a 'site-packages' folder.

    :param path: The site-packages folder into which to install the package.
    :param package: The name of the package in the wheel. Defaults to ``dummy``
    :param version: The version number of the package. Defaults to ``1.2.3``
    :param tag: The installation tag for the package. Defaults to a pure python wheel.
    :param extra_content: Optional. A list of tuples of ``(path, content)`` or
        ``(path, content, chmod)`` that will be added to the wheel. If ``chmod`` is
        not specified, default filesystem permissions will be used.
    """
    for entry in installed_package_content(
        package=package,
        version=version,
        tag=tag,
        extra_content=extra_content,
    ):
        try:
            filename, content, chmod = entry
        except ValueError:
            filename, content = entry
            chmod = None
        mode = "wb" if isinstance(content, bytes) else "w"
        create_file(path / filename, content=content, mode=mode, chmod=chmod)


def create_wheel(
    path,
    package="dummy",
    version="1.2.3",
    tag="py3-none-any",
    extra_content=None,
):
    """Create a sample wheel file.

    :param path: The folder where the wheel should be written.
    :param package: The name of the package in the wheel. Defaults to ``dummy``
    :param version: The version number of the package. Defaults to ``1.2.3``
    :param tag: The installation tag for the package. Defaults to a pure python wheel.
    :param extra_content: Optional. A list of tuples of ``(path, content)`` or
        ``(path, content, chmod)`` that will be added to the wheel. If ``chmod`` is
        not specified, default filesystem permissions will be used.
    """
    wheel_filename = path / f"{package}-{version}-{tag}.whl"

    create_zip_file(
        wheel_filename,
        content=installed_package_content(
            package=package,
            version=version,
            tag=tag,
            extra_content=extra_content,
        ),
    )

    return wheel_filename


def file_content(path: Path) -> str | bytes | None:
    """Return the content of a file, or None if the path is a directory."""
    if path.is_dir():
        return None
    try:
        with path.open(encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError:
        with path.open(mode="rb") as f:
            return f.read()


def assert_url_resolvable(url: str):
    """Tests whether a URL is resolvable with retries; raises for failure."""
    transport = RetryTransport(
        # this retry for the underlying transport only applies to connection attempts
        transport=httpx.HTTPTransport(retries=3),
        # the underlying retry timing algorithm (first retry is immediate):
        #  > backoff_factor * (2^attempt) * random.uniform(jitter, 1)
        # Here are some example backoff timings using the defaults below:
        #   [0.0, 0.694, 1.939, 3.588]
        #   [0.0, 0.529, 1.589, 2.002]
        #   [0.0, 0.599, 1.948, 4.184]
        retry=Retry(
            total=3,
            backoff_factor=0.6,
            backoff_jitter=0.3,
            allowed_methods=[
                "HEAD",
                "GET",
                "PUT",
                "DELETE",
                "OPTIONS",
                "TRACE",
            ],
            status_forcelist=[
                HTTPStatus.TOO_MANY_REQUESTS,
                HTTPStatus.BAD_GATEWAY,
                HTTPStatus.SERVICE_UNAVAILABLE,
                HTTPStatus.GATEWAY_TIMEOUT,
            ],
            retry_on_exceptions=[
                httpx.TimeoutException,
                httpx.NetworkError,
                httpx.RemoteProtocolError,
                httpx.ReadTimeout,
            ],
        ),
    )

    try:
        with httpx.Client(transport=transport, follow_redirects=True) as client:
            response = client.head(url, timeout=10)
        response.raise_for_status()
    finally:
        # RetryTransport doesn't close its transport...so close it manually
        transport._sync_transport.close()



import re

import pytest


def test_boolean_question_yes(console):
    """Test that boolean_question returns True when user selects Yes."""
    console.values = ["y"]

    result = console.boolean_question(
        description="Confirm?",
        intro="Are you sure?",
        default=None,
    )

    assert console.prompts == ["Confirm? [y/n]? "]
    assert result is True


def test_boolean_question_no(console):
    """Test that boolean_question returns False when user selects No."""
    console.values = ["n"]

    result = console.boolean_question(
        description="Confirm?",
        intro="Are you sure?",
        default=None,
    )

    assert console.prompts == ["Confirm? [y/n]? "]
    assert result is False


@pytest.mark.parametrize(
    ("default", "expected", "prompt"),
    [
        (True, True, "Confirm? [Y/n]? "),
        (False, False, "Confirm? [y/N]? "),
    ],
)
def test_boolean_question_default_used(console, default, expected, prompt):
    """If no input is provided, the default value should be used."""
    console.values = [""]

    result = console.boolean_question(
        description="Confirm?",
        intro="Are you sure?",
        default=default,
    )

    assert console.prompts == [prompt]
    assert result == expected


@pytest.mark.parametrize(
    ("override_value", "expected_result"),
    [
        ("Yes", True),
        ("no", False),
    ],
)
def test_boolean_question_override_used(
    console,
    capsys,
    override_value,
    expected_result,
):
    """The override is used if provided and valid (parametrized for True/False
    cases)."""
    result = console.boolean_question(
        description="Confirm?",
        intro="Are you sure?",
        default=None,
        override_value=override_value,
    )

    output = capsys.readouterr().out
    assert f"Using override value {override_value!r}" in output
    assert result == expected_result
    assert console.prompts == []


def test_boolean_question_override_invalid(console):
    """If override_value is invalid, an error should be raised immediately."""
    with pytest.raises(
        ValueError,
        match=re.escape("Invalid boolean value: 'Yeah Nah'. Expected one of "),
    ):
        console.boolean_question(
            description="Confirm?",
            intro="Are you sure?",
            default=True,
            override_value="Yeah Nah",
        )


def test_boolean_question_empty_input_no_default_returns_false(console):
    """Reprompts the user when no override or default given."""
    console.values = ["", "", "n"]  # Simulates pressing Enter

    result = console.boolean_question(
        description="Confirm?",
        intro="Are you sure?",
        default=None,
        override_value=None,
    )

    assert result is False



import json
import platform
import subprocess
from signal import SIGTERM
from unittest import mock

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS import macOS_log_clean_filter
from briefcase.platforms.macOS.app import macOSAppRunCommand


@pytest.fixture
def run_command(dummy_console, tmp_path):
    command = macOSAppRunCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command._stream_app_logs = mock.MagicMock()

    # To satisfy coverage, the stop function must be invoked
    # at least once when streaming app logs.
    def mock_stream_app_logs(app, stop_func=lambda: 1 + 1, **kwargs):
        stop_func()

    command._stream_app_logs.side_effect = mock_stream_app_logs
    command.tools.os.kill = mock.MagicMock()

    return command


def test_run_gui_app(run_command, first_app_config, sleep_zero, tmp_path, monkeypatch):
    """A macOS GUI app can be started."""
    # Mock a popen object that represents the log stream
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = log_stream_process

    # Monkeypatch the tools get the process ID
    monkeypatch.setattr(
        "briefcase.platforms.macOS.get_process_id_by_command", lambda *a, **kw: 100
    )

    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path],
        cwd=tmp_path / "home",
        check=True,
    )

    # The log stream was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The app process was killed on exit.
    run_command.tools.os.kill.assert_called_with(100, SIGTERM)


def test_run_gui_app_with_passthrough(
    run_command,
    first_app_config,
    sleep_zero,
    tmp_path,
    monkeypatch,
):
    """A macOS app can be started in debug mode with args."""
    run_command.console.verbosity = LogLevel.DEBUG

    # Mock a popen object that represents the log stream
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = log_stream_process

    # Monkeypatch the tools get the process ID
    monkeypatch.setattr(
        "briefcase.platforms.macOS.get_process_id_by_command", lambda *a, **kw: 100
    )

    # Run the app with args
    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path, "--args", "foo", "--bar"],
        cwd=tmp_path / "home",
        check=True,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # The log stream was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The app process was killed on exit.
    run_command.tools.os.kill.assert_called_with(100, SIGTERM)


def test_run_gui_app_failed(run_command, first_app_config, sleep_zero, tmp_path):
    """If there's a problem started the app, an exception is raised."""
    # Mock a failure opening the app
    run_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["open", "-n", run_command.binary_path(first_app_config)],
        returncode=1,
    )

    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path],
        cwd=tmp_path / "home",
        check=True,
    )

    # No attempt was made to stream the log or cleanup
    run_command._stream_app_logs.assert_not_called()
    run_command.tools.os.kill.assert_not_called()


def test_run_gui_app_find_pid_failed(
    run_command,
    first_app_config,
    sleep_zero,
    tmp_path,
    monkeypatch,
    capsys,
):
    """If after app is started, its pid is not found, do not stream output."""
    # Mock a failed PID lookup
    monkeypatch.setattr(
        "briefcase.platforms.macOS.get_process_id_by_command",
        lambda *a, **kw: None,
    )

    with pytest.raises(BriefcaseCommandError) as exc_info:
        run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path],
        cwd=tmp_path / "home",
        check=True,
    )
    assert exc_info.value.msg == (
        "Unable to find process for app first-app to start log streaming."
    )

    # No attempt was made to stream the log or cleanup
    run_command._stream_app_logs.assert_not_called()
    run_command.tools.os.kill.assert_not_called()


def test_run_gui_app_test_mode(
    run_command,
    first_app_config,
    sleep_zero,
    tmp_path,
    monkeypatch,
):
    """A macOS GUI app can be started in test mode."""
    first_app_config.test_mode = True

    # Mock a popen object that represents the log stream
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = log_stream_process

    # Monkeypatch the tools get the process ID
    monkeypatch.setattr(
        "briefcase.platforms.macOS.get_process_id_by_command", lambda *a, **kw: 100
    )

    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path],
        cwd=tmp_path / "home",
        check=True,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The log stream was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The app process was killed on exit.
    run_command.tools.os.kill.assert_called_with(100, SIGTERM)


def test_run_gui_app_debugger(
    run_command,
    first_app_config,
    sleep_zero,
    tmp_path,
    monkeypatch,
    dummy_debugger,
):
    """A macOS GUI app can be started in debug mode."""
    # Mock a popen object that represents the log stream
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = log_stream_process

    first_app_config.debugger = dummy_debugger
    first_app_config.debugger_host = "somehost"
    first_app_config.debugger_port = 9999

    # Monkeypatch the tools get the process ID
    monkeypatch.setattr(
        "briefcase.platforms.macOS.get_process_id_by_command", lambda *a, **kw: 100
    )

    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    sender = bin_path / "Contents/MacOS/First App"
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            "log",
            "stream",
            "--style",
            "compact",
            "--predicate",
            f'senderImagePath=="{sender}"'
            f' OR (processImagePath=="{sender}"'
            ' AND senderImagePath=="/usr/lib/libffi.dylib")',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )
    run_command.tools.subprocess.run.assert_called_with(
        ["open", "-n", bin_path],
        cwd=tmp_path / "home",
        check=True,
        env={
            "BRIEFCASE_DEBUGGER": json.dumps(
                {
                    "debugger": "dummy",
                    "host": "somehost",
                    "port": 9999,
                    "host_os": platform.system(),
                    "app_path_mappings": {
                        "device_sys_path_regex": "app$",
                        "device_subfolders": ["first_app"],
                        "host_folders": [str(tmp_path / "base_path/src/first_app")],
                    },
                    "app_packages_path_mappings": None,
                }
            )
        },
    )

    # The log stream was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The app process was killed on exit.
    run_command.tools.os.kill.assert_called_with(100, SIGTERM)


def test_run_console_app(run_command, first_app_config, tmp_path):
    """A macOS console app can be started."""
    # Set the app to be a console app
    first_app_config.console_app = True

    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    run_command.tools.subprocess.run.assert_called_with(
        [bin_path / "Contents/MacOS/First App"],
        cwd=tmp_path / "home",
        check=True,
        stream_output=False,
    )

    # The log stream was not started
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_with_passthrough(
    run_command,
    first_app_config,
    tmp_path,
):
    """A macOS console app can be started in debug mode with args."""
    run_command.console.verbosity = LogLevel.DEBUG

    # Set the app to be a console app
    first_app_config.console_app = True

    # Run the app with args
    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    run_command.tools.subprocess.run.assert_called_with(
        [bin_path / "Contents/MacOS/First App", "foo", "--bar"],
        cwd=tmp_path / "home",
        check=True,
        stream_output=False,
        env={"BRIEFCASE_DEBUG": "1"},
    )

    # The log stream was not started
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_test_mode(run_command, first_app_config, sleep_zero, tmp_path):
    """A macOS console app can be started in test mode."""
    first_app_config.console_app = True
    first_app_config.test_mode = True

    # Mock a popen object that represents the app
    app_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = app_process

    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    run_command.tools.subprocess.Popen.assert_called_with(
        [bin_path / "Contents/MacOS/First App"],
        cwd=tmp_path / "home",
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The log stream was not started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=app_process,
    )


def test_run_console_app_test_mode_with_passthrough(
    run_command,
    first_app_config,
    sleep_zero,
    tmp_path,
):
    """A macOS console app can be started in test mode with parameters and debug
    mode."""
    run_command.console.verbosity = LogLevel.DEBUG

    first_app_config.console_app = True
    first_app_config.test_mode = True

    # Mock a popen object that represents the app
    app_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.return_value = app_process

    run_command.run_app(
        first_app_config,
        passthrough=["foo", "--bar"],
    )

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    run_command.tools.subprocess.Popen.assert_called_with(
        [bin_path / "Contents/MacOS/First App", "foo", "--bar"],
        cwd=tmp_path / "home",
        env={"BRIEFCASE_DEBUG": "1", "BRIEFCASE_MAIN_MODULE": "tests.first_app"},
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The log stream was not started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=app_process,
    )


def test_run_console_app_failed(run_command, first_app_config, sleep_zero, tmp_path):
    """If there's a problem started a console app, an exception is raised."""
    # Set the app to be a console app
    first_app_config.console_app = True

    # Mock a failure opening the app
    run_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=[run_command.binary_path(first_app_config) / "Contents/MacOS/First App"],
        returncode=1,
    )

    # Although the command raises an error, this could be because the script itself
    # raised an error.
    run_command.run_app(first_app_config, passthrough=[])

    # Calls were made to start the app and to start a log stream.
    bin_path = run_command.binary_path(first_app_config)
    run_command.tools.subprocess.run.assert_called_with(
        [bin_path / "Contents/MacOS/First App"],
        cwd=tmp_path / "home",
        stream_output=False,
        check=True,
    )

    # No attempt was made to stream the log or cleanup
    run_command._stream_app_logs.assert_not_called()



import zipfile
from pathlib import Path, PurePosixPath

import pytest

from briefcase.platforms.web.static import StaticWebBuildCommand


@pytest.fixture
def build_command(dummy_console, tmp_path):
    return StaticWebBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_handle_legacy_css_warn_and_append(
    build_command,
    tmp_path,
    capsys,
):
    """Legacy CSS warns and appends to new contrib key."""
    wheel_path = Path(tmp_path) / "dummy-1.2.3.whl"
    with zipfile.ZipFile(wheel_path, "w") as zf:
        zf.writestr("dummy/static/one.css", "h1 { x:1; }")

    inserts = {}
    with zipfile.ZipFile(wheel_path) as zf:
        # First call - warn
        build_command._handle_legacy_css(
            wheel=zf,
            path=PurePosixPath("dummy/static/one.css"),
            package_key="pkg 1.0",
            inserts=inserts,
        )
        # Pre-seed same contrib key to force append on second call
        target_map = inserts.setdefault("static/css/style.css", {}).setdefault(
            "css", {}
        )
        key = "pkg 1.0 (legacy static CSS: one.css)"
        target_map[key] = target_map[key] + "\n/*extra*/"
        # Second call - no additional warning, content appended
        build_command._handle_legacy_css(
            wheel=zf,
            path=PurePosixPath("dummy/static/one.css"),
            package_key="pkg 1.0",
            inserts=inserts,
        )

    # One warning for each file
    output = capsys.readouterr().out
    assert (
        "dummy-1.2.3.whl: legacy '/static' CSS file dummy/static/one.css detected."
        in output
    )

    # Content appended
    out = inserts["static/css/style.css"]["css"][key]
    assert "h1 { x:1; }" in out
    assert "/*extra*/" in out


def test_handle_legacy_css_non_utf8_raise(build_command, tmp_path):
    """Non-UTF8 legacy CSS raises an UnicodeDecodeError."""
    bad = b"\xff\xfe\xfa"
    wheel_path = Path(tmp_path) / "dummy-1.2.3.whl"
    with zipfile.ZipFile(wheel_path, "w") as zf:
        zf.writestr("dummy/static/bad.css", bad)

    inserts = {}
    with zipfile.ZipFile(wheel_path) as zf, pytest.raises(UnicodeDecodeError):
        build_command._handle_legacy_css(
            wheel=zf,
            path=PurePosixPath("dummy/static/bad.css"),
            package_key="pkg 1.0",
            inserts=inserts,
        )
